from contextlib import contextmanager

import pytest

from dagster import (
    AssetKey,
    AssetMaterialization,
    DagsterEventType,
    Output,
    execute_pipeline,
    pipeline,
    seven,
    solid,
)
from dagster.core.events.log import EventRecord
from dagster.core.instance import DagsterInstance, InstanceType
from dagster.core.launcher.sync_in_memory_run_launcher import SyncInMemoryRunLauncher
from dagster.core.storage.event_log import (
    ConsolidatedSqliteEventLogStorage,
    InMemoryEventLogStorage,
)
from dagster.core.storage.noop_compute_log_manager import NoOpComputeLogManager
from dagster.core.storage.root import LocalArtifactStorage
from dagster.core.storage.runs import InMemoryRunStorage


def get_instance(temp_dir, event_log_storage):
    return DagsterInstance(
        instance_type=InstanceType.EPHEMERAL,
        local_artifact_storage=LocalArtifactStorage(temp_dir),
        run_storage=InMemoryRunStorage(),
        event_storage=event_log_storage,
        compute_log_manager=NoOpComputeLogManager(),
        run_launcher=SyncInMemoryRunLauncher(),
    )


@contextmanager
def create_in_memory_event_log_instance():
    with seven.TemporaryDirectory() as temp_dir:
        asset_storage = InMemoryEventLogStorage()
        instance = get_instance(temp_dir, asset_storage)
        yield [instance, asset_storage]


@contextmanager
def create_consolidated_sqlite_event_log_instance():
    with seven.TemporaryDirectory() as temp_dir:
        asset_storage = ConsolidatedSqliteEventLogStorage(temp_dir)
        instance = get_instance(temp_dir, asset_storage)
        yield [instance, asset_storage]


asset_test = pytest.mark.parametrize(
    'asset_aware_context',
    [create_in_memory_event_log_instance, create_consolidated_sqlite_event_log_instance,],
)


@solid
def solid_one(_):
    yield AssetMaterialization(asset_key=AssetKey('asset_1'))
    yield Output(1)


@solid
def solid_two(_):
    yield AssetMaterialization(asset_key=AssetKey('asset_2'))
    yield AssetMaterialization(asset_key=AssetKey(['path', 'to', 'asset_3']))
    yield Output(1)


@solid
def solid_normalization(_):
    yield AssetMaterialization(asset_key='path/to-asset_4')
    yield Output(1)


@pipeline
def pipeline_one():
    solid_one()


@pipeline
def pipeline_two():
    solid_one()
    solid_two()


@pipeline
def pipeline_normalization():
    solid_normalization()


@asset_test
def test_asset_keys(asset_aware_context):
    with asset_aware_context() as ctx:
        instance, event_log_storage = ctx
        execute_pipeline(pipeline_one, instance=instance)
        execute_pipeline(pipeline_two, instance=instance)
        asset_keys = event_log_storage.get_all_asset_keys()
        assert len(asset_keys) == 3
        assert set([asset_key.to_string() for asset_key in asset_keys]) == set(
            ['asset_1', 'asset_2', 'path.to.asset_3']
        )


@asset_test
def test_asset_events(asset_aware_context):
    with asset_aware_context() as ctx:
        instance, event_log_storage = ctx
        execute_pipeline(pipeline_one, instance=instance)
        execute_pipeline(pipeline_two, instance=instance)
        asset_events = event_log_storage.get_asset_events(AssetKey('asset_1'))
        assert len(asset_events) == 2
        for event in asset_events:
            assert isinstance(event, EventRecord)
            assert event.is_dagster_event
            assert event.dagster_event.event_type == DagsterEventType.STEP_MATERIALIZATION
            assert event.dagster_event.asset_key

        asset_events = event_log_storage.get_asset_events(AssetKey(['path', 'to', 'asset_3']))
        assert len(asset_events) == 1


@asset_test
def test_asset_run_ids(asset_aware_context):
    with asset_aware_context() as ctx:
        instance, event_log_storage = ctx
        one = execute_pipeline(pipeline_one, instance=instance)
        two = execute_pipeline(pipeline_two, instance=instance)
        run_ids = event_log_storage.get_asset_run_ids(AssetKey('asset_1'))
        assert set(run_ids) == set([one.run_id, two.run_id])


@asset_test
def test_asset_normalization(asset_aware_context):
    with asset_aware_context() as ctx:
        instance, event_log_storage = ctx
        execute_pipeline(pipeline_normalization, instance=instance)
        asset_keys = event_log_storage.get_all_asset_keys()
        assert len(asset_keys) == 1
        asset_key = asset_keys[0]
        assert asset_key.to_string() == 'path.to.asset_4'
        assert asset_key.path == ['path', 'to', 'asset_4']


@asset_test
def test_asset_wipe(asset_aware_context):
    with asset_aware_context() as ctx:
        instance, event_log_storage = ctx
        one = execute_pipeline(pipeline_one, instance=instance)
        execute_pipeline(pipeline_two, instance=instance)
        asset_keys = event_log_storage.get_all_asset_keys()
        assert len(asset_keys) == 3
        log_count = len(event_log_storage.get_logs_for_run(one.run_id))
        instance.wipe_assets(asset_keys)
        asset_keys = event_log_storage.get_all_asset_keys()
        assert len(asset_keys) == 0
        assert log_count == len(event_log_storage.get_logs_for_run(one.run_id))

        execute_pipeline(pipeline_one, instance=instance)
        execute_pipeline(pipeline_two, instance=instance)
        asset_keys = event_log_storage.get_all_asset_keys()
        assert len(asset_keys) == 3
        instance.wipe_assets([AssetKey(['path', 'to', 'asset_3'])])
        asset_keys = event_log_storage.get_all_asset_keys()
        assert len(asset_keys) == 2



import inspect
from collections import namedtuple

from dagster import DagsterInvariantViolationError, PipelineDefinition, RepositoryDefinition
from dagster.core.code_pointer import load_python_file, load_python_module

LoadableTarget = namedtuple('LoadableTarget', 'attribute target_definition')


def loadable_targets_from_python_file(python_file, working_directory=None):
    loaded_module = load_python_file(python_file, working_directory)
    return loadable_targets_from_loaded_module(loaded_module)


def loadable_targets_from_python_module(module_name, remove_from_path_fn=None):
    module = load_python_module(
        module_name, warn_only=True, remove_from_path_fn=remove_from_path_fn
    )
    return loadable_targets_from_loaded_module(module)


def loadable_targets_from_python_package(package_name, remove_from_path_fn=None):
    module = load_python_module(package_name, remove_from_path_fn=remove_from_path_fn)
    return loadable_targets_from_loaded_module(module)


def loadable_targets_from_loaded_module(module):
    loadable_repos = _loadable_targets_of_type(module, RepositoryDefinition)
    if loadable_repos:
        return loadable_repos

    loadable_pipelines = _loadable_targets_of_type(module, PipelineDefinition)

    if not loadable_pipelines:
        raise DagsterInvariantViolationError(
            'No pipelines or repositories found in "{}".'.format(module.__name__)
        )

    elif len(loadable_pipelines) > 1:
        raise DagsterInvariantViolationError(
            (
                'No repository and more than one pipeline found in "{module_name}". If you load '
                'a file or module directly it must either have one repository or one '
                'pipeline in scope. Found pipelines defined in variables or decorated '
                'functions: {pipeline_symbols}.'
            ).format(
                module_name=module.__name__,
                pipeline_symbols=repr([p.attribute for p in loadable_pipelines]),
            )
        )

    else:
        return loadable_pipelines


def _loadable_targets_of_type(module, klass):
    loadable_targets = []
    for name, value in inspect.getmembers(module):
        if isinstance(value, klass):
            loadable_targets.append(LoadableTarget(name, value))

    return loadable_targets



import pytest

from dagster import (
    DagsterInvalidDefinitionError,
    ModeDefinition,
    ResourceDefinition,
    execute_pipeline,
    pipeline,
    system_storage,
)
from dagster.core.storage.system_storage import create_mem_system_storage_data


def test_naked_system_storage():
    called = {}

    @system_storage(required_resource_keys=set())
    def no_config_storage(init_context):
        called['called'] = True
        return create_mem_system_storage_data(init_context)

    @pipeline(mode_defs=[ModeDefinition(system_storage_defs=[no_config_storage])])
    def pass_pipeline():
        pass

    assert execute_pipeline(
        pass_pipeline, run_config={'storage': {'no_config_storage': {}}}
    ).success

    assert called['called']

    # This also works with None because storage is a selector and you can indicate your
    # "selection" by the presence of the key
    assert execute_pipeline(
        pass_pipeline, run_config={'storage': {'no_config_storage': None}}
    ).success


def test_resource_requirements_pass():
    called = {}

    @system_storage(required_resource_keys={'yup'})
    def storage_with_req(init_context):
        assert hasattr(init_context.resources, 'yup')
        assert not hasattr(init_context.resources, 'not_required')
        assert not hasattr(init_context.resources, 'kjdkfjdkfje')
        called['called'] = True
        return create_mem_system_storage_data(init_context)

    @pipeline(
        mode_defs=[
            ModeDefinition(
                resource_defs={
                    'yup': ResourceDefinition.none_resource(),
                    'not_required': ResourceDefinition.none_resource(),
                },
                system_storage_defs=[storage_with_req],
            )
        ]
    )
    def resource_req_pass_pipeline():
        pass

    assert execute_pipeline(
        resource_req_pass_pipeline, run_config={'storage': {'storage_with_req': {}}}
    ).success

    assert called['called']

    # This also works with None because storage is a selector and you can indicate your
    # "selection" by the presence of the key
    assert execute_pipeline(
        resource_req_pass_pipeline, run_config={'storage': {'storage_with_req': None}}
    ).success


def test_resource_requirements_fail():
    @system_storage(required_resource_keys={'yup'})
    def storage_with_req(init_context):
        return create_mem_system_storage_data(init_context)

    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:

        @pipeline(
            mode_defs=[
                ModeDefinition(
                    resource_defs={'nope': ResourceDefinition.none_resource()},
                    system_storage_defs=[storage_with_req],
                )
            ]
        )
        def _resource_req_pass_pipeline():
            pass

    assert str(exc_info.value) == (
        'Resource \'yup\' is required by system storage \'storage_with_req\', but '
        'is not provided by mode \'default\'.'
    )



import sys
from collections import namedtuple
from enum import Enum

import pytest

from dagster.check import CheckError, ParameterCheckError, inst_param, set_param
from dagster.serdes import (
    Persistable,
    SerdesClassUsageError,
    _deserialize_json_to_dagster_namedtuple,
    _pack_value,
    _serialize_dagster_namedtuple,
    _unpack_value,
    _whitelist_for_persistence,
    _whitelist_for_serdes,
    default_to_storage_value,
    deserialize_json_to_dagster_namedtuple,
    deserialize_value,
    serialize_value,
)
from dagster.utils import compose


def test_deserialize_value_ok():
    unpacked_tuple = deserialize_value('{"foo": "bar"}')
    assert unpacked_tuple
    assert unpacked_tuple['foo'] == 'bar'


def test_deserialize_json_to_dagster_namedtuple_non_namedtuple():
    with pytest.raises(CheckError):
        deserialize_json_to_dagster_namedtuple('{"foo": "bar"}')


@pytest.mark.parametrize('bad_obj', [1, None, False])
def test_deserialize_json_to_dagster_namedtuple_invalid_types(bad_obj):
    with pytest.raises(ParameterCheckError):
        deserialize_json_to_dagster_namedtuple(bad_obj)


def test_deserialize_empty_set():
    assert set() == deserialize_value(serialize_value(set()))
    assert frozenset() == deserialize_value(serialize_value(frozenset()))


def _initial_whitelist_map():
    return {
        'types': {'tuple': {}, 'enum': {}},
        'persistence': {},
    }


def test_forward_compat_serdes_new_field_with_default():
    _TEST_WHITELIST_MAP = _initial_whitelist_map()

    @_whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP)
    class Quux(namedtuple('_Quux', 'foo bar')):
        def __new__(cls, foo, bar):
            return super(Quux, cls).__new__(cls, foo, bar)  # pylint: disable=bad-super-call

    assert 'Quux' in _TEST_WHITELIST_MAP['types']['tuple']
    assert _TEST_WHITELIST_MAP['types']['tuple']['Quux'] == Quux

    quux = Quux('zip', 'zow')

    serialized = _serialize_dagster_namedtuple(quux, whitelist_map=_TEST_WHITELIST_MAP)

    # pylint: disable=function-redefined
    @_whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP)
    class Quux(namedtuple('_Quux', 'foo bar baz')):  # pylint: disable=bad-super-call
        def __new__(cls, foo, bar, baz=None):
            return super(Quux, cls).__new__(cls, foo, bar, baz=baz)

    assert 'Quux' in _TEST_WHITELIST_MAP['types']['tuple']
    assert _TEST_WHITELIST_MAP['types']['tuple']['Quux'] == Quux

    deserialized = _deserialize_json_to_dagster_namedtuple(
        serialized, whitelist_map=_TEST_WHITELIST_MAP
    )

    assert deserialized != quux
    assert deserialized.foo == quux.foo
    assert deserialized.bar == quux.bar
    assert deserialized.baz is None


def test_forward_compat_serdes_new_enum_field():
    _TEST_WHITELIST_MAP = _initial_whitelist_map()

    @_whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP)
    class Corge(Enum):
        FOO = 1
        BAR = 2

    assert 'Corge' in _TEST_WHITELIST_MAP['types']['enum']

    corge = Corge.FOO

    packed = _pack_value(corge, whitelist_map=_TEST_WHITELIST_MAP)

    # pylint: disable=function-redefined
    @_whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP)
    class Corge(Enum):
        FOO = 1
        BAR = 2
        BAZ = 3

    unpacked = _unpack_value(packed, whitelist_map=_TEST_WHITELIST_MAP)

    assert unpacked != corge
    assert unpacked.name == corge.name
    assert unpacked.value == corge.value


# This behavior isn't possible on 2.7 because of `inspect` limitations
@pytest.mark.skipif(sys.version_info < (3,), reason="This behavior isn't available on 2.7")
def test_backward_compat_serdes():
    _TEST_WHITELIST_MAP = _initial_whitelist_map()

    @_whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP)
    class Quux(namedtuple('_Quux', 'foo bar baz')):
        def __new__(cls, foo, bar, baz):
            return super(Quux, cls).__new__(cls, foo, bar, baz)  # pylint: disable=bad-super-call

    quux = Quux('zip', 'zow', 'whoopie')

    serialized = _serialize_dagster_namedtuple(quux, whitelist_map=_TEST_WHITELIST_MAP)

    # pylint: disable=function-redefined
    @_whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP)
    class Quux(namedtuple('_Quux', 'foo bar')):  # pylint: disable=bad-super-call
        def __new__(cls, foo, bar):
            return super(Quux, cls).__new__(cls, foo, bar)

    deserialized = _deserialize_json_to_dagster_namedtuple(
        serialized, whitelist_map=_TEST_WHITELIST_MAP
    )

    assert deserialized != quux
    assert deserialized.foo == quux.foo
    assert deserialized.bar == quux.bar
    assert not hasattr(deserialized, 'baz')


def serdes_test_class(klass):
    _TEST_WHITELIST_MAP = _initial_whitelist_map()

    return _whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP)(klass)


@pytest.mark.skipif(
    sys.version_info.major < 3, reason='Serdes declaration time checks python 3 only'
)
def test_wrong_first_arg():
    with pytest.raises(SerdesClassUsageError) as exc_info:

        @serdes_test_class
        class NotCls(namedtuple('NotCls', 'field_one field_two')):
            def __new__(not_cls, field_two, field_one):
                return super(NotCls, not_cls).__new__(field_one, field_two)

    assert (
        str(exc_info.value)
        == 'For namedtuple NotCls: First parameter must be _cls or cls. Got "not_cls".'
    )


@pytest.mark.skipif(
    sys.version_info.major < 3, reason='Serdes declaration time checks python 3 only'
)
def test_incorrect_order():
    with pytest.raises(SerdesClassUsageError) as exc_info:

        @serdes_test_class
        class WrongOrder(namedtuple('WrongOrder', 'field_one field_two')):
            def __new__(cls, field_two, field_one):
                return super(WrongOrder, cls).__new__(field_one, field_two)

    assert str(exc_info.value) == (
        'For namedtuple WrongOrder: '
        'Params to __new__ must match the order of field declaration '
        'in the namedtuple. Declared field number 1 in the namedtuple '
        'is "field_one". Parameter 1 in __new__ method is "field_two".'
    )


@pytest.mark.skipif(
    sys.version_info.major < 3, reason='Serdes declaration time checks python 3 only'
)
def test_missing_one_parameter():
    with pytest.raises(SerdesClassUsageError) as exc_info:

        @serdes_test_class
        class MissingFieldInNew(namedtuple('MissingFieldInNew', 'field_one field_two field_three')):
            def __new__(cls, field_one, field_two):
                return super(MissingFieldInNew, cls).__new__(field_one, field_two, None)

    assert str(exc_info.value) == (
        "For namedtuple MissingFieldInNew: "
        "Missing parameters to __new__. You have declared fields in "
        "the named tuple that are not present as parameters to the "
        "to the __new__ method. In order for both serdes serialization "
        "and pickling to work, these must match. Missing: ['field_three']"
    )


@pytest.mark.skipif(
    sys.version_info.major < 3, reason='Serdes declaration time checks python 3 only'
)
def test_missing_many_parameters():
    with pytest.raises(SerdesClassUsageError) as exc_info:

        @serdes_test_class
        class MissingFieldsInNew(
            namedtuple('MissingFieldsInNew', 'field_one field_two field_three, field_four')
        ):
            def __new__(cls, field_one, field_two):
                return super(MissingFieldsInNew, cls).__new__(field_one, field_two, None, None)

    assert str(exc_info.value) == (
        "For namedtuple MissingFieldsInNew: "
        "Missing parameters to __new__. You have declared fields in "
        "the named tuple that are not present as parameters to the "
        "to the __new__ method. In order for both serdes serialization "
        "and pickling to work, these must match. Missing: ['field_three', 'field_four']"
    )


@pytest.mark.skipif(
    sys.version_info.major < 3, reason='Serdes declaration time checks python 3 only'
)
def test_extra_parameters_must_have_defaults():
    with pytest.raises(SerdesClassUsageError) as exc_info:

        @serdes_test_class
        class OldFieldsWithoutDefaults(
            namedtuple('OldFieldsWithoutDefaults', 'field_three field_four')
        ):
            # pylint:disable=unused-argument
            def __new__(
                cls,
                field_three,
                field_four,
                # Graveyard Below
                field_one,
                field_two,
            ):
                return super(OldFieldsWithoutDefaults, cls).__new__(field_three, field_four)

    assert str(exc_info.value) == (
        'For namedtuple OldFieldsWithoutDefaults: '
        'Parameter "field_one" is a parameter to the __new__ '
        'method but is not a field in this namedtuple. '
        'The only reason why this should exist is that '
        'it is a field that used to exist (we refer to '
        'this as the graveyard) but no longer does. However '
        'it might exist in historical storage. This parameter '
        'existing ensures that serdes continues to work. However '
        'these must come at the end and have a default value for pickling to work.'
    )


@pytest.mark.skipif(
    sys.version_info.major < 3, reason='Serdes declaration time checks python 3 only'
)
def test_extra_parameters_have_working_defaults():
    @serdes_test_class
    class OldFieldsWithDefaults(namedtuple('OldFieldsWithDefaults', 'field_three field_four')):
        # pylint:disable=unused-argument
        def __new__(
            cls,
            field_three,
            field_four,
            # Graveyard Below
            none_field=None,
            falsey_field=0,
            another_falsey_field='',
            value_field='klsjkfjd',
        ):
            return super(OldFieldsWithDefaults, cls).__new__(field_three, field_four)


def test_set():
    _TEST_WHITELIST_MAP = _initial_whitelist_map()

    @_whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP)
    class HasSets(namedtuple('_HasSets', 'reg_set frozen_set')):
        def __new__(cls, reg_set, frozen_set):
            set_param(reg_set, 'reg_set')
            inst_param(frozen_set, 'frozen_set', frozenset)
            return super(HasSets, cls).__new__(cls, reg_set, frozen_set)

    foo = HasSets({1, 2, 3, '3'}, frozenset([4, 5, 6, '6']))

    serialized = _serialize_dagster_namedtuple(foo, whitelist_map=_TEST_WHITELIST_MAP)
    foo_2 = _deserialize_json_to_dagster_namedtuple(serialized, whitelist_map=_TEST_WHITELIST_MAP)
    assert foo == foo_2


def test_persistent_tuple():
    _TEST_WHITELIST_MAP = _initial_whitelist_map()

    def whitelist_for_persistence(klass):
        return compose(
            _whitelist_for_persistence(whitelist_map=_TEST_WHITELIST_MAP),
            _whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP),
        )(klass)

    @whitelist_for_persistence
    class Alphabet(namedtuple('_Alphabet', 'a b c'), Persistable):
        def __new__(cls, a, b, c):
            return super(Alphabet, cls).__new__(cls, a, b, c)

    foo = Alphabet(a='A', b='B', c='C')
    serialized = _serialize_dagster_namedtuple(foo, whitelist_map=_TEST_WHITELIST_MAP)
    foo_2 = _deserialize_json_to_dagster_namedtuple(serialized, whitelist_map=_TEST_WHITELIST_MAP)
    assert foo == foo_2


def test_from_storage_dict():
    _TEST_WHITELIST_MAP = _initial_whitelist_map()

    def whitelist_for_persistence(klass):
        return compose(
            _whitelist_for_persistence(whitelist_map=_TEST_WHITELIST_MAP),
            _whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP),
        )(klass)

    @whitelist_for_persistence
    class DeprecatedAlphabet(namedtuple('_DeprecatedAlphabet', 'a b c'), Persistable):
        def __new__(cls, a, b, c):
            raise Exception('DeprecatedAlphabet is deprecated')

        @classmethod
        def from_storage_dict(cls, storage_dict):
            # instead of the DeprecatedAlphabet, directly invoke the namedtuple constructor
            return super(DeprecatedAlphabet, cls).__new__(
                cls, storage_dict.get('a'), storage_dict.get('b'), storage_dict.get('c'),
            )

    assert 'DeprecatedAlphabet' in _TEST_WHITELIST_MAP['persistence']
    serialized = '{"__class__": "DeprecatedAlphabet", "a": "A", "b": "B", "c": "C"}'

    nt = _deserialize_json_to_dagster_namedtuple(serialized, whitelist_map=_TEST_WHITELIST_MAP)
    assert isinstance(nt, DeprecatedAlphabet)


def test_to_storage_value():
    _TEST_WHITELIST_MAP = _initial_whitelist_map()

    def whitelist_for_persistence(klass):
        return compose(
            _whitelist_for_persistence(whitelist_map=_TEST_WHITELIST_MAP),
            _whitelist_for_serdes(whitelist_map=_TEST_WHITELIST_MAP),
        )(klass)

    @whitelist_for_persistence
    class DeprecatedAlphabet(namedtuple('_DeprecatedAlphabet', 'a b c'), Persistable):
        def __new__(cls, a, b, c):
            return super(DeprecatedAlphabet, cls).__new__(cls, a, b, c)

        def to_storage_value(self):
            return default_to_storage_value(
                SubstituteAlphabet(self.a, self.b, self.c), _TEST_WHITELIST_MAP
            )

    @whitelist_for_persistence
    class SubstituteAlphabet(namedtuple('_SubstituteAlphabet', 'a b c'), Persistable):
        def __new__(cls, a, b, c):
            return super(SubstituteAlphabet, cls).__new__(cls, a, b, c)

    nested = DeprecatedAlphabet(None, None, '_C')
    deprecated = DeprecatedAlphabet('A', 'B', nested)
    serialized = _serialize_dagster_namedtuple(deprecated, whitelist_map=_TEST_WHITELIST_MAP)
    alphabet = _deserialize_json_to_dagster_namedtuple(
        serialized, whitelist_map=_TEST_WHITELIST_MAP
    )
    assert not isinstance(alphabet, DeprecatedAlphabet)
    assert isinstance(alphabet, SubstituteAlphabet)
    assert not isinstance(alphabet.c, DeprecatedAlphabet)
    assert isinstance(alphabet.c, SubstituteAlphabet)



import pytest

from dagster import Field, execute_solid, solid


def test_default_implies_not_required_field_correct():
    @solid(config_schema={'default_to_one': Field(int, default_value=1)})
    def return_default_to_one(context):
        return context.solid_config['default_to_one']

    default_to_one_field = return_default_to_one.config_field.config_type.fields['default_to_one']
    assert default_to_one_field.is_required is False


def test_default_implies_not_required_execute_solid():
    @solid(config_schema={'default_to_one': Field(int, default_value=1)})
    def return_default_to_one(context):
        return context.solid_config['default_to_one']

    # This currently throws. It should not require configuration as the
    # the default value should make it *not* required.
    execute_solid(return_default_to_one)



import string

from dagster.api.snapshot_partition import (
    sync_get_external_partition_config,
    sync_get_external_partition_config_grpc,
    sync_get_external_partition_names,
    sync_get_external_partition_names_grpc,
    sync_get_external_partition_set_execution_param_data,
    sync_get_external_partition_set_execution_param_data_grpc,
    sync_get_external_partition_tags,
    sync_get_external_partition_tags_grpc,
)
from dagster.core.host_representation import (
    ExternalPartitionConfigData,
    ExternalPartitionExecutionErrorData,
    ExternalPartitionNamesData,
    ExternalPartitionSetExecutionParamData,
    ExternalPartitionTagsData,
)

from .utils import get_bar_grpc_repo_handle, get_bar_repo_handle


def test_external_partition_names():
    repository_handle = get_bar_repo_handle()
    data = sync_get_external_partition_names(repository_handle, 'baz_partitions')
    assert isinstance(data, ExternalPartitionNamesData)
    assert data.partition_names == list(string.ascii_lowercase)


def test_external_partition_names_error():
    repository_handle = get_bar_repo_handle()
    error = sync_get_external_partition_names(repository_handle, 'error_partitions')
    assert isinstance(error, ExternalPartitionExecutionErrorData)
    assert 'womp womp' in error.error.to_string()


def test_external_partition_names_grpc():
    repository_handle = get_bar_grpc_repo_handle()
    data = sync_get_external_partition_names_grpc(
        repository_handle.repository_location_handle.client, repository_handle, 'baz_partitions'
    )
    assert isinstance(data, ExternalPartitionNamesData)
    assert data.partition_names == list(string.ascii_lowercase)


def test_external_partition_names_error_grpc():
    repository_handle = get_bar_grpc_repo_handle()

    error = sync_get_external_partition_names_grpc(
        repository_handle.repository_location_handle.client, repository_handle, 'error_partitions'
    )
    assert isinstance(error, ExternalPartitionExecutionErrorData)
    assert 'womp womp' in error.error.to_string()


def test_external_partitions_config():
    repository_handle = get_bar_repo_handle()
    data = sync_get_external_partition_config(repository_handle, 'baz_partitions', 'c')
    assert isinstance(data, ExternalPartitionConfigData)
    assert data.run_config
    assert data.run_config['solids']['do_input']['inputs']['x']['value'] == 'c'


def test_external_partitions_config_error():
    repository_handle = get_bar_repo_handle()
    error = sync_get_external_partition_config(repository_handle, 'error_partition_config', 'c')
    assert isinstance(error, ExternalPartitionExecutionErrorData)


def test_external_partitions_config_grpc():
    repository_handle = get_bar_grpc_repo_handle()
    data = sync_get_external_partition_config_grpc(
        repository_handle.repository_location_handle.client,
        repository_handle,
        'baz_partitions',
        'c',
    )
    assert isinstance(data, ExternalPartitionConfigData)
    assert data.run_config
    assert data.run_config['solids']['do_input']['inputs']['x']['value'] == 'c'


def test_external_partitions_config_error_grpc():
    repository_handle = get_bar_grpc_repo_handle()
    error = sync_get_external_partition_config_grpc(
        repository_handle.repository_location_handle.client,
        repository_handle,
        'error_partition_config',
        'c',
    )
    assert isinstance(error, ExternalPartitionExecutionErrorData)


def test_external_partitions_tags():
    repository_handle = get_bar_repo_handle()
    data = sync_get_external_partition_tags(repository_handle, 'baz_partitions', 'c')
    assert isinstance(data, ExternalPartitionTagsData)
    assert data.tags
    assert data.tags['foo'] == 'bar'


def test_external_partitions_tags_error():
    repository_handle = get_bar_repo_handle()
    error = sync_get_external_partition_tags(repository_handle, 'error_partition_tags', 'c')
    assert isinstance(error, ExternalPartitionExecutionErrorData)


def test_external_partitions_tags_grpc():
    repository_handle = get_bar_grpc_repo_handle()
    data = sync_get_external_partition_tags_grpc(
        repository_handle.repository_location_handle.client,
        repository_handle,
        'baz_partitions',
        'c',
    )
    assert isinstance(data, ExternalPartitionTagsData)
    assert data.tags
    assert data.tags['foo'] == 'bar'


def test_external_partitions_tags_error_grpc():
    repository_handle = get_bar_grpc_repo_handle()
    error = sync_get_external_partition_tags_grpc(
        repository_handle.repository_location_handle.client,
        repository_handle,
        'error_partition_tags',
        'c',
    )
    assert isinstance(error, ExternalPartitionExecutionErrorData)


def test_external_partition_set_execution_params():
    repository_handle = get_bar_repo_handle()
    data = sync_get_external_partition_set_execution_param_data(
        repository_handle, 'baz_partitions', ['a', 'b', 'c']
    )
    assert isinstance(data, ExternalPartitionSetExecutionParamData)
    assert len(data.partition_data) == 3


def test_external_partition_set_execution_params_grpc():
    repository_handle = get_bar_grpc_repo_handle()
    data = sync_get_external_partition_set_execution_param_data_grpc(
        repository_handle.repository_location_handle.client,
        repository_handle,
        'baz_partitions',
        ['a', 'b', 'c'],
    )
    assert isinstance(data, ExternalPartitionSetExecutionParamData)
    assert len(data.partition_data) == 3


def test_external_partition_set_execution_params_error():
    repository_handle = get_bar_repo_handle()
    error = sync_get_external_partition_set_execution_param_data(
        repository_handle, 'error_partitions', ['a', 'b', 'c']
    )
    assert isinstance(error, ExternalPartitionExecutionErrorData)
    assert 'womp womp' in error.error.to_string()



from dagster import pipeline, solid


def test_solid_tags():
    @solid(tags={'foo': 'bar'})
    def tags_solid(_):
        pass

    assert tags_solid.tags == {'foo': 'bar'}

    @solid()
    def no_tags_solid(_):
        pass

    assert no_tags_solid.tags == {}


def test_pipeline_tags():
    @pipeline(tags={'foo': 'bar'})
    def tags_pipeline():
        pass

    assert tags_pipeline.tags == {'foo': 'bar'}

    @pipeline()
    def no_tags_pipeline():
        pass

    assert no_tags_pipeline.tags == {}


def test_solid_subset_tags():
    @solid
    def noop_solid(_):
        pass

    @pipeline(tags={'foo': 'bar'})
    def tags_pipeline():
        noop_solid()

    assert tags_pipeline.get_pipeline_subset_def({'noop_solid'}).tags == {'foo': 'bar'}

    @pipeline()
    def no_tags_pipeline():
        noop_solid()

    assert no_tags_pipeline.get_pipeline_subset_def({'noop_solid'}).tags == {}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['test_basic_solids_config 1'] = {
    'execution': {
        'in_process': {
            'config': {'marker_to_close': '', 'retries': {'disabled': {}, 'enabled': {}}}
        },
        'multiprocess': {
            'config': {'max_concurrent': 0, 'retries': {'disabled': {}, 'enabled': {}}}
        },
    },
    'loggers': {'console': {'config': {'log_level': '', 'name': ''}}},
    'resources': {},
    'solids': {'required_field_solid': {'config': {'required_int': 0}}},
    'storage': {'filesystem': {'config': {'base_dir': ''}}, 'in_memory': {}},
    'intermediate_storage': {'filesystem': {'config': {'base_dir': ''}}, 'in_memory': {}},
}

snapshots['test_two_modes 1'] = {'resources': {'value': {'config': {'mode_one_field': ''}}}}

snapshots['test_two_modes 2'] = {
    'execution': {
        'in_process': {
            'config': {'marker_to_close': '', 'retries': {'disabled': {}, 'enabled': {}}}
        },
        'multiprocess': {
            'config': {'max_concurrent': 0, 'retries': {'disabled': {}, 'enabled': {}}}
        },
    },
    'loggers': {'console': {'config': {'log_level': '', 'name': ''}}},
    'resources': {'value': {'config': {'mode_one_field': ''}}},
    'solids': {},
    'storage': {'filesystem': {'config': {'base_dir': ''}}, 'in_memory': {}},
    'intermediate_storage': {'filesystem': {'config': {'base_dir': ''}}, 'in_memory': {}},
}

snapshots['test_two_modes 3'] = {'resources': {'value': {'config': {'mode_two_field': 0}}}}

snapshots['test_two_modes 4'] = {
    'execution': {
        'in_process': {
            'config': {'marker_to_close': '', 'retries': {'disabled': {}, 'enabled': {}}}
        },
        'multiprocess': {
            'config': {'max_concurrent': 0, 'retries': {'disabled': {}, 'enabled': {}}}
        },
    },
    'loggers': {'console': {'config': {'log_level': '', 'name': ''}}},
    'resources': {'value': {'config': {'mode_two_field': 0}}},
    'solids': {},
    'storage': {'filesystem': {'config': {'base_dir': ''}}, 'in_memory': {}},
    'intermediate_storage': {'filesystem': {'config': {'base_dir': ''}}, 'in_memory': {}},
}



import os

import pytest


@pytest.fixture(scope='session', autouse=True)
def unset_dagster_home():
    old_env = os.getenv('DAGSTER_HOME')
    if old_env is not None:
        del os.environ['DAGSTER_HOME']
    yield
    if old_env is not None:
        os.environ['DAGSTER_HOME'] = old_env



import os
import shutil
import uuid
from collections import defaultdict
from contextlib import contextmanager

# top-level include is dangerous in terms of incurring circular deps
from dagster import (
    DagsterInvariantViolationError,
    DependencyDefinition,
    Failure,
    ModeDefinition,
    PipelineDefinition,
    RepositoryDefinition,
    SolidInvocation,
    SystemStorageData,
    TypeCheck,
    check,
    execute_pipeline,
    lambda_solid,
    seven,
)
from dagster.core.definitions.executable import InMemoryExecutablePipeline
from dagster.core.definitions.logger import LoggerDefinition
from dagster.core.definitions.resource import ScopedResourcesBuilder
from dagster.core.definitions.solid import ISolidDefinition
from dagster.core.execution.api import create_execution_plan, scoped_pipeline_context
from dagster.core.execution.context_creation_pipeline import (
    SystemPipelineExecutionContext,
    construct_execution_context_data,
    create_context_creation_data,
    create_executor,
    create_log_manager,
)
from dagster.core.instance import DagsterInstance
from dagster.core.scheduler import ScheduleStatus, Scheduler
from dagster.core.scheduler.scheduler import DagsterScheduleDoesNotExist, DagsterSchedulerError
from dagster.core.snap import snapshot_from_execution_plan
from dagster.core.storage.file_manager import LocalFileManager
from dagster.core.storage.intermediate_storage import InMemoryIntermediateStorage
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.types.dagster_type import resolve_dagster_type
from dagster.core.utility_solids import define_stub_solid
from dagster.core.utils import make_new_run_id
from dagster.serdes import ConfigurableClass

# pylint: disable=unused-import
from ..temp_file import (
    get_temp_dir,
    get_temp_file_handle,
    get_temp_file_handle_with_data,
    get_temp_file_name,
    get_temp_file_name_with_data,
    get_temp_file_names,
)
from ..typing_api import is_typing_type


def create_test_pipeline_execution_context(logger_defs=None):
    from dagster.core.storage.intermediate_storage import build_in_mem_intermediates_storage

    loggers = check.opt_dict_param(
        logger_defs, 'logger_defs', key_type=str, value_type=LoggerDefinition
    )
    mode_def = ModeDefinition(logger_defs=loggers)
    pipeline_def = PipelineDefinition(
        name='test_legacy_context', solid_defs=[], mode_defs=[mode_def]
    )
    run_config = {'loggers': {key: {} for key in loggers}}
    pipeline_run = PipelineRun(pipeline_name='test_legacy_context', run_config=run_config)
    instance = DagsterInstance.ephemeral()
    execution_plan = create_execution_plan(pipeline=pipeline_def, run_config=run_config)
    creation_data = create_context_creation_data(execution_plan, run_config, pipeline_run, instance)
    log_manager = create_log_manager(creation_data)
    scoped_resources_builder = ScopedResourcesBuilder()
    executor = create_executor(creation_data)

    return SystemPipelineExecutionContext(
        construct_execution_context_data(
            context_creation_data=creation_data,
            scoped_resources_builder=scoped_resources_builder,
            intermediate_storage=build_in_mem_intermediates_storage(pipeline_run.run_id),
            system_storage_data=SystemStorageData(
                intermediate_storage=InMemoryIntermediateStorage(),
                file_manager=LocalFileManager.for_instance(instance, pipeline_run.run_id),
            ),
            log_manager=log_manager,
            retries=executor.retries,
            raise_on_error=True,
        ),
        executor=executor,
        log_manager=log_manager,
    )


def _dep_key_of(solid):
    return SolidInvocation(solid.definition.name, solid.name)


def build_pipeline_with_input_stubs(pipeline_def, inputs):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.dict_param(inputs, 'inputs', key_type=str, value_type=dict)

    deps = defaultdict(dict)
    for solid_name, dep_dict in pipeline_def.dependencies.items():
        for input_name, dep in dep_dict.items():
            deps[solid_name][input_name] = dep

    stub_solid_defs = []

    for solid_name, input_dict in inputs.items():
        if not pipeline_def.has_solid_named(solid_name):
            raise DagsterInvariantViolationError(
                (
                    'You are injecting an input value for solid {solid_name} '
                    'into pipeline {pipeline_name} but that solid was not found'
                ).format(solid_name=solid_name, pipeline_name=pipeline_def.name)
            )

        solid = pipeline_def.solid_named(solid_name)
        for input_name, input_value in input_dict.items():
            stub_solid_def = define_stub_solid(
                '__stub_{solid_name}_{input_name}'.format(
                    solid_name=solid_name, input_name=input_name
                ),
                input_value,
            )
            stub_solid_defs.append(stub_solid_def)
            deps[_dep_key_of(solid)][input_name] = DependencyDefinition(stub_solid_def.name)

    return PipelineDefinition(
        name=pipeline_def.name + '_stubbed',
        solid_defs=pipeline_def.top_level_solid_defs + stub_solid_defs,
        mode_defs=pipeline_def.mode_definitions,
        dependencies=deps,
    )


def execute_solids_within_pipeline(
    pipeline_def,
    solid_names,
    inputs=None,
    run_config=None,
    mode=None,
    preset=None,
    tags=None,
    instance=None,
):
    '''Execute a set of solids within an existing pipeline.

    Intended to support tests. Input values may be passed directly.

    Args:
        pipeline_def (PipelineDefinition): The pipeline within which to execute the solid.
        solid_names (FrozenSet[str]): A set of the solid names, or the aliased solids, to execute.
        inputs (Optional[Dict[str, Dict[str, Any]]]): A dict keyed on solid names, whose values are
            dicts of input names to input values, used to pass input values to the solids directly.
            You may also use the ``run_config`` to configure any inputs that are configurable.
        run_config (Optional[dict]): The environment configuration that parameterized this
            execution, as a dict.
        mode (Optional[str]): The name of the pipeline mode to use. You may not set both ``mode``
            and ``preset``.
        preset (Optional[str]): The name of the pipeline preset to use. You may not set both
            ``mode`` and ``preset``.
        tags (Optional[Dict[str, Any]]): Arbitrary key-value pairs that will be added to pipeline
            logs.
        instance (Optional[DagsterInstance]): The instance to execute against. If this is ``None``,
            an ephemeral instance will be used, and no artifacts will be persisted from the run.

    Returns:
        Dict[str, Union[CompositeSolidExecutionResult, SolidExecutionResult]]: The results of
        executing the solids, keyed by solid name.
    '''
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.set_param(solid_names, 'solid_names', of_type=str)
    inputs = check.opt_dict_param(inputs, 'inputs', key_type=str, value_type=dict)

    sub_pipeline = pipeline_def.get_pipeline_subset_def(solid_names)
    stubbed_pipeline = build_pipeline_with_input_stubs(sub_pipeline, inputs)
    result = execute_pipeline(
        stubbed_pipeline,
        run_config=run_config,
        mode=mode,
        preset=preset,
        tags=tags,
        instance=instance,
    )

    return {sr.solid.name: sr for sr in result.solid_result_list}


def execute_solid_within_pipeline(
    pipeline_def,
    solid_name,
    inputs=None,
    run_config=None,
    mode=None,
    preset=None,
    tags=None,
    instance=None,
):
    '''Execute a single solid within an existing pipeline.

    Intended to support tests. Input values may be passed directly.

    Args:
        pipeline_def (PipelineDefinition): The pipeline within which to execute the solid.
        solid_name (str): The name of the solid, or the aliased solid, to execute.
        inputs (Optional[Dict[str, Any]]): A dict of input names to input values, used to
            pass input values to the solid directly. You may also use the ``run_config`` to
            configure any inputs that are configurable.
        run_config (Optional[dict]): The environment configuration that parameterized this
            execution, as a dict.
        mode (Optional[str]): The name of the pipeline mode to use. You may not set both ``mode``
            and ``preset``.
        preset (Optional[str]): The name of the pipeline preset to use. You may not set both
            ``mode`` and ``preset``.
        tags (Optional[Dict[str, Any]]): Arbitrary key-value pairs that will be added to pipeline
            logs.
        instance (Optional[DagsterInstance]): The instance to execute against. If this is ``None``,
            an ephemeral instance will be used, and no artifacts will be persisted from the run.

    Returns:
        Union[CompositeSolidExecutionResult, SolidExecutionResult]: The result of executing the
        solid.
    '''

    return execute_solids_within_pipeline(
        pipeline_def,
        solid_names={solid_name},
        inputs={solid_name: inputs} if inputs else None,
        run_config=run_config,
        mode=mode,
        preset=preset,
        tags=tags,
        instance=instance,
    )[solid_name]


@contextmanager
def yield_empty_pipeline_context(run_id=None, instance=None):
    pipeline = InMemoryExecutablePipeline(PipelineDefinition([]))
    pipeline_def = pipeline.get_definition()
    instance = check.opt_inst_param(
        instance, 'instance', DagsterInstance, default=DagsterInstance.ephemeral()
    )

    execution_plan = create_execution_plan(pipeline)

    pipeline_run = instance.create_run(
        pipeline_name='<empty>',
        run_id=run_id,
        run_config=None,
        mode=None,
        solids_to_execute=None,
        step_keys_to_execute=None,
        status=None,
        tags=None,
        root_run_id=None,
        parent_run_id=None,
        pipeline_snapshot=pipeline_def.get_pipeline_snapshot(),
        execution_plan_snapshot=snapshot_from_execution_plan(
            execution_plan, pipeline_def.get_pipeline_snapshot_id()
        ),
        parent_pipeline_snapshot=pipeline_def.get_parent_pipeline_snapshot(),
    )
    with scoped_pipeline_context(execution_plan, {}, pipeline_run, instance) as context:
        yield context


def execute_solid(
    solid_def, mode_def=None, input_values=None, tags=None, run_config=None, raise_on_error=True,
):
    '''Execute a single solid in an ephemeral pipeline.

    Intended to support unit tests. Input values may be passed directly, and no pipeline need be
    specified -- an ephemeral pipeline will be constructed.

    Args:
        solid_def (SolidDefinition): The solid to execute.
        mode_def (Optional[ModeDefinition]): The mode within which to execute the solid. Use this
            if, e.g., custom resources, loggers, or executors are desired.
        input_values (Optional[Dict[str, Any]]): A dict of input names to input values, used to
            pass inputs to the solid directly. You may also use the ``run_config`` to
            configure any inputs that are configurable.
        tags (Optional[Dict[str, Any]]): Arbitrary key-value pairs that will be added to pipeline
            logs.
        run_config (Optional[dict]): The environment configuration that parameterized this
            execution, as a dict.
        raise_on_error (Optional[bool]): Whether or not to raise exceptions when they occur.
            Defaults to ``True``, since this is the most useful behavior in test.

    Returns:
        Union[CompositeSolidExecutionResult, SolidExecutionResult]: The result of executing the
        solid.
    '''
    check.inst_param(solid_def, 'solid_def', ISolidDefinition)
    check.opt_inst_param(mode_def, 'mode_def', ModeDefinition)
    input_values = check.opt_dict_param(input_values, 'input_values', key_type=str)
    solid_defs = [solid_def]

    def create_value_solid(input_name, input_value):
        @lambda_solid(name=input_name)
        def input_solid():
            return input_value

        return input_solid

    dependencies = defaultdict(dict)

    for input_name, input_value in input_values.items():
        dependencies[solid_def.name][input_name] = DependencyDefinition(input_name)
        solid_defs.append(create_value_solid(input_name, input_value))

    result = execute_pipeline(
        PipelineDefinition(
            name='ephemeral_{}_solid_pipeline'.format(solid_def.name),
            solid_defs=solid_defs,
            dependencies=dependencies,
            mode_defs=[mode_def] if mode_def else None,
        ),
        run_config=run_config,
        mode=mode_def.name if mode_def else None,
        tags=tags,
        raise_on_error=raise_on_error,
    )
    return result.result_for_handle(solid_def.name)


def check_dagster_type(dagster_type, value):
    '''Test a custom Dagster type.

    Args:
        dagster_type (Any): The Dagster type to test. Should be one of the
            :ref:`built-in types <builtin>`, a dagster type explicitly constructed with
            :py:func:`as_dagster_type`, :py:func:`@usable_as_dagster_type <dagster_type>`, or
            :py:func:`PythonObjectDagsterType`, or a Python type.
        value (Any): The runtime value to test.

    Returns:
        TypeCheck: The result of the type check.


    Examples:

        .. code-block:: python

            assert check_dagster_type(Dict[Any, Any], {'foo': 'bar'}).success
    '''

    if is_typing_type(dagster_type):
        raise DagsterInvariantViolationError(
            (
                'Must pass in a type from dagster module. You passed {dagster_type} '
                'which is part of python\'s typing module.'
            ).format(dagster_type=dagster_type)
        )

    dagster_type = resolve_dagster_type(dagster_type)
    with yield_empty_pipeline_context() as pipeline_context:
        context = pipeline_context.for_type(dagster_type)
        try:
            type_check = dagster_type.type_check(context, value)
        except Failure as failure:
            return TypeCheck(success=False, description=failure.description)

        if not isinstance(type_check, TypeCheck):
            raise DagsterInvariantViolationError(
                'Type checks can only return TypeCheck. Type {type_name} returned {value}.'.format(
                    type_name=dagster_type.name, value=repr(type_check)
                )
            )
        return type_check


@contextmanager
def restore_directory(src):
    with seven.TemporaryDirectory() as temp_dir:
        dst = os.path.join(temp_dir, os.path.basename(src))
        shutil.copytree(src, dst)
        try:
            yield
        finally:
            shutil.rmtree(src)
            shutil.copytree(dst, src)


class FilesystemTestScheduler(Scheduler, ConfigurableClass):
    '''This class is used in dagster core and dagster_graphql to test the scheduler's interactions
    with schedule storage, which are implemented in the methods defined on the base Scheduler class.
    Therefore, the following methods used to actually schedule jobs (e.g. create and remove cron jobs
    on a cron tab) are left unimplemented.
    '''

    def __init__(self, artifacts_dir, inst_data=None):
        check.str_param(artifacts_dir, 'artifacts_dir')
        self._artifacts_dir = artifacts_dir
        self._inst_data = inst_data

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {'base_dir': str}

    @staticmethod
    def from_config_value(inst_data, config_value):
        return FilesystemTestScheduler(artifacts_dir=config_value['base_dir'], inst_data=inst_data)

    def debug_info(self):
        return ""

    def start_schedule(self, instance, external_schedule):
        pass

    def stop_schedule(self, instance, schedule_origin_id):
        pass

    def running_schedule_count(self, schedule_origin_id):
        return 0

    def get_logs_path(self, _instance, schedule_origin_id):
        check.str_param(schedule_origin_id, 'schedule_origin_id')
        return os.path.join(self._artifacts_dir, 'logs', schedule_origin_id, 'scheduler.log')

    def wipe(self, instance):
        pass



from dagster import check
from dagster.core.definitions import (
    AssetMaterialization,
    ExpectationResult,
    Materialization,
    Output,
    Solid,
    SolidHandle,
)
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.execution.context.compute import SolidExecutionContext
from dagster.core.execution.context.system import SystemComputeExecutionContext

from .objects import ExecutionStep, StepInput, StepKind, StepOutput


def create_compute_step(pipeline_name, environment_config, solid, step_inputs, handle):
    check.str_param(pipeline_name, 'pipeline_name')
    check.inst_param(solid, 'solid', Solid)
    check.list_param(step_inputs, 'step_inputs', of_type=StepInput)
    check.opt_inst_param(handle, 'handle', SolidHandle)

    # the environment config has the solid output name configured
    config_output_names = set()
    current_handle = handle
    while current_handle:
        solid_config = environment_config.solids.get(current_handle.to_string())
        current_handle = current_handle.parent
        for output_spec in solid_config.outputs:
            config_output_names = config_output_names.union(output_spec.keys())

    return ExecutionStep(
        pipeline_name=pipeline_name,
        key_suffix='compute',
        step_inputs=step_inputs,
        step_outputs=[
            StepOutput(
                name=name,
                dagster_type=output_def.dagster_type,
                optional=output_def.optional,
                should_materialize=name in config_output_names,
            )
            for name, output_def in solid.definition.output_dict.items()
        ],
        compute_fn=lambda step_context, inputs: _execute_core_compute(
            step_context.for_compute(), inputs, solid.definition.compute_fn
        ),
        kind=StepKind.COMPUTE,
        solid_handle=handle,
        solid=solid,
        tags=solid.tags,
        hook_defs=solid.hook_defs,
    )


def _yield_compute_results(compute_context, inputs, compute_fn):
    check.inst_param(compute_context, 'compute_context', SystemComputeExecutionContext)
    step = compute_context.step
    user_event_sequence = compute_fn(SolidExecutionContext(compute_context), inputs)

    if isinstance(user_event_sequence, Output):
        raise DagsterInvariantViolationError(
            (
                'Compute function for solid {solid_name} returned a Output rather than '
                'yielding it. The compute_fn of the core SolidDefinition must yield '
                'its results'
            ).format(solid_name=str(step.solid_handle))
        )

    if user_event_sequence is None:
        return

    for event in user_event_sequence:
        if isinstance(event, (Output, AssetMaterialization, Materialization, ExpectationResult)):
            yield event
        else:
            raise DagsterInvariantViolationError(
                (
                    'Compute function for solid {solid_name} yielded a value of type {type_} '
                    'rather than an instance of Output, AssetMaterialization, or ExpectationResult.'
                    ' Values yielded by solids must be wrapped in one of these types. If your '
                    'solid has a single output and yields no other events, you may want to use '
                    '`return` instead of `yield` in the body of your solid compute function. If '
                    'you are already using `return`, and you expected to return a value of type '
                    '{type_}, you may be inadvertently returning a generator rather than the value '
                    'you expected.'
                ).format(solid_name=str(step.solid_handle), type_=type(event))
            )


def _execute_core_compute(compute_context, inputs, compute_fn):
    '''
    Execute the user-specified compute for the solid. Wrap in an error boundary and do
    all relevant logging and metrics tracking
    '''
    check.inst_param(compute_context, 'compute_context', SystemComputeExecutionContext)
    check.dict_param(inputs, 'inputs', key_type=str)

    step = compute_context.step

    all_results = []
    for step_output in _yield_compute_results(compute_context, inputs, compute_fn):
        yield step_output
        if isinstance(step_output, Output):
            all_results.append(step_output)

    if len(all_results) != len(step.step_outputs):
        emitted_result_names = {r.output_name for r in all_results}
        solid_output_names = {output.name for output in step.step_outputs}
        omitted_outputs = solid_output_names.difference(emitted_result_names)
        compute_context.log.info(
            'Solid {solid} did not fire outputs {outputs}'.format(
                solid=str(step.solid_handle), outputs=repr(omitted_outputs)
            )
        )



from collections import namedtuple

from dagster import check
from dagster.core.code_pointer import FileCodePointer, ModuleCodePointer, PackageCodePointer
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.origin import RepositoryPythonOrigin


class LoadableTargetOrigin(
    namedtuple(
        'LoadableTargetOrigin',
        'executable_path python_file module_name working_directory attribute',
    )
):
    def __new__(
        cls,
        executable_path=None,
        python_file=None,
        module_name=None,
        working_directory=None,
        attribute=None,
    ):
        return super(LoadableTargetOrigin, cls).__new__(
            cls,
            executable_path=check.opt_str_param(executable_path, 'executable_path'),
            python_file=check.opt_str_param(python_file, 'python_file'),
            module_name=check.opt_str_param(module_name, 'module_name'),
            working_directory=check.opt_str_param(working_directory, 'working_directory'),
            attribute=check.opt_str_param(attribute, 'attribute'),
        )

    @staticmethod
    def from_python_origin(repository_python_origin):
        check.inst_param(
            repository_python_origin, 'repository_python_origin', RepositoryPythonOrigin
        )
        executable_path = repository_python_origin.executable_path
        code_pointer = repository_python_origin.code_pointer
        if isinstance(code_pointer, FileCodePointer):
            return LoadableTargetOrigin(
                executable_path=executable_path,
                python_file=code_pointer.python_file,
                attribute=code_pointer.fn_name,
                working_directory=code_pointer.working_directory,
            )
        elif isinstance(code_pointer, ModuleCodePointer):
            return LoadableTargetOrigin(
                executable_path=executable_path,
                module_name=code_pointer.module,
                attribute=code_pointer.fn_name,
            )
        elif isinstance(code_pointer, PackageCodePointer):
            return LoadableTargetOrigin(
                executable_path=executable_path,
                module_name=code_pointer.module,
                attribute=code_pointer.attribute,
            )
        else:
            raise DagsterInvariantViolationError(
                "Unexpected code pointer {code_pointer_name}".format(
                    code_pointer_name=type(code_pointer).__name__
                )
            )



import re

import pytest

from dagster.api.snapshot_execution_plan import (
    sync_get_external_execution_plan,
    sync_get_external_execution_plan_grpc,
)
from dagster.core.errors import DagsterSubprocessError
from dagster.core.snap.execution_plan_snapshot import ExecutionPlanSnapshot

from .utils import get_foo_grpc_pipeline_handle, get_foo_pipeline_handle


def test_execution_plan_snapshot_api():
    pipeline_handle = get_foo_pipeline_handle()

    execution_plan_snapshot = sync_get_external_execution_plan(
        pipeline_handle.get_origin(), run_config={}, mode="default", pipeline_snapshot_id="12345",
    )

    assert isinstance(execution_plan_snapshot, ExecutionPlanSnapshot)
    assert execution_plan_snapshot.step_keys_to_execute == [
        'do_something.compute',
        'do_input.compute',
    ]
    assert len(execution_plan_snapshot.steps) == 2


def test_execution_plan_error_api():
    pipeline_handle = get_foo_pipeline_handle()

    with pytest.raises(
        DagsterSubprocessError, match=re.escape('Could not find mode made_up_mode in pipeline foo')
    ):
        sync_get_external_execution_plan(
            pipeline_handle.get_origin(),
            run_config={},
            mode='made_up_mode',
            pipeline_snapshot_id='12345',
        )


def test_execution_plan_error_grpc():
    pipeline_handle = get_foo_grpc_pipeline_handle()
    api_client = pipeline_handle.repository_handle.repository_location_handle.client

    with pytest.raises(
        DagsterSubprocessError, match=re.escape('Could not find mode made_up_mode in pipeline foo')
    ):
        sync_get_external_execution_plan_grpc(
            api_client,
            pipeline_handle.get_origin(),
            run_config={},
            mode='made_up_mode',
            pipeline_snapshot_id='12345',
        )


def test_execution_plan_with_step_keys_to_execute_snapshot_api():
    pipeline_handle = get_foo_pipeline_handle()

    execution_plan_snapshot = sync_get_external_execution_plan(
        pipeline_handle.get_origin(),
        run_config={},
        mode='default',
        pipeline_snapshot_id='12345',
        step_keys_to_execute=['do_something.compute'],
    )

    assert isinstance(execution_plan_snapshot, ExecutionPlanSnapshot)
    assert execution_plan_snapshot.step_keys_to_execute == [
        'do_something.compute',
    ]
    assert len(execution_plan_snapshot.steps) == 2


def test_execution_plan_with_subset_snapshot_api():
    pipeline_handle = get_foo_pipeline_handle()

    execution_plan_snapshot = sync_get_external_execution_plan(
        pipeline_handle.get_origin(),
        run_config={'solids': {'do_input': {'inputs': {'x': {'value': "test"}}}}},
        mode="default",
        pipeline_snapshot_id="12345",
        solid_selection=["do_input"],
    )

    assert isinstance(execution_plan_snapshot, ExecutionPlanSnapshot)
    assert execution_plan_snapshot.step_keys_to_execute == [
        'do_input.compute',
    ]
    assert len(execution_plan_snapshot.steps) == 1


def test_execution_plan_snapshot_api_grpc():
    pipeline_handle = get_foo_grpc_pipeline_handle()
    api_client = pipeline_handle.repository_handle.repository_location_handle.client

    execution_plan_snapshot = sync_get_external_execution_plan_grpc(
        api_client,
        pipeline_handle.get_origin(),
        run_config={},
        mode="default",
        pipeline_snapshot_id="12345",
    )

    assert isinstance(execution_plan_snapshot, ExecutionPlanSnapshot)
    assert execution_plan_snapshot.step_keys_to_execute == [
        'do_something.compute',
        'do_input.compute',
    ]
    assert len(execution_plan_snapshot.steps) == 2


def test_execution_plan_with_step_keys_to_execute_snapshot_api_grpc():
    pipeline_handle = get_foo_grpc_pipeline_handle()
    api_client = pipeline_handle.repository_handle.repository_location_handle.client

    execution_plan_snapshot = sync_get_external_execution_plan_grpc(
        api_client,
        pipeline_handle.get_origin(),
        run_config={},
        mode="default",
        pipeline_snapshot_id="12345",
        step_keys_to_execute=['do_something.compute'],
    )

    assert isinstance(execution_plan_snapshot, ExecutionPlanSnapshot)
    assert execution_plan_snapshot.step_keys_to_execute == [
        'do_something.compute',
    ]
    assert len(execution_plan_snapshot.steps) == 2


def test_execution_plan_with_subset_snapshot_api_grpc():
    pipeline_handle = get_foo_grpc_pipeline_handle()
    api_client = pipeline_handle.repository_handle.repository_location_handle.client

    execution_plan_snapshot = sync_get_external_execution_plan_grpc(
        api_client,
        pipeline_handle.get_origin(),
        run_config={'solids': {'do_input': {'inputs': {'x': {'value': "test"}}}}},
        mode="default",
        pipeline_snapshot_id="12345",
        solid_selection=["do_input"],
    )

    assert isinstance(execution_plan_snapshot, ExecutionPlanSnapshot)
    assert execution_plan_snapshot.step_keys_to_execute == [
        'do_input.compute',
    ]
    assert len(execution_plan_snapshot.steps) == 1



from dagster import InputDefinition, lambda_solid, pipeline
from dagster.core.selector.subset_selector import (
    MAX_NUM,
    Traverser,
    generate_dep_graph,
    parse_clause,
    parse_solid_selection,
)


@lambda_solid
def return_one():
    return 1


@lambda_solid
def return_two():
    return 2


@lambda_solid(input_defs=[InputDefinition('num1'), InputDefinition('num2')])
def add_nums(num1, num2):
    return num1 + num2


@lambda_solid(input_defs=[InputDefinition('num')])
def multiply_two(num):
    return num * 2


@lambda_solid(input_defs=[InputDefinition('num')])
def add_one(num):
    return num + 1


@pipeline
def foo_pipeline():
    '''
    return_one ---> add_nums --> multiply_two --> add_one
    return_two --|
    '''
    return add_one(multiply_two(add_nums(return_one(), return_two())))


def test_generate_dep_graph():
    graph = generate_dep_graph(foo_pipeline)
    assert graph == {
        'upstream': {
            'return_one': set(),
            'return_two': set(),
            'add_nums': {'return_one', 'return_two'},
            'multiply_two': {'add_nums'},
            'add_one': {'multiply_two'},
        },
        'downstream': {
            'return_one': {'add_nums'},
            'return_two': {'add_nums'},
            'add_nums': {'multiply_two'},
            'multiply_two': {'add_one'},
            'add_one': set(),
        },
    }


def test_traverser():
    graph = generate_dep_graph(foo_pipeline)
    traverser = Traverser(graph)

    assert traverser.fetch_upstream(item_name='return_one', depth=1) == set()
    assert traverser.fetch_downstream(item_name='return_one', depth=1) == {'add_nums'}
    assert traverser.fetch_upstream(item_name='multiply_two', depth=0) == set()
    assert traverser.fetch_upstream(item_name='multiply_two', depth=2) == {
        'add_nums',
        'return_one',
        'return_two',
    }
    assert traverser.fetch_downstream(item_name='multiply_two', depth=2) == {'add_one'}


def test_traverser_invalid():
    graph = generate_dep_graph(foo_pipeline)
    traverser = Traverser(graph)

    assert traverser.fetch_upstream(item_name='some_solid', depth=1) == set()


def test_parse_clause():
    assert parse_clause('some_solid') == (0, 'some_solid', 0)
    assert parse_clause('*some_solid') == (MAX_NUM, 'some_solid', 0)
    assert parse_clause('some_solid+') == (0, 'some_solid', 1)
    assert parse_clause('+some_solid+') == (1, 'some_solid', 1)
    assert parse_clause('*some_solid++') == (MAX_NUM, 'some_solid', 2)


def test_parse_clause_invalid():
    assert parse_clause('1+some_solid') == None


def test_parse_solid_selection_single():
    solid_selection_single = parse_solid_selection(foo_pipeline, ['add_nums'])
    assert len(solid_selection_single) == 1
    assert solid_selection_single == {'add_nums'}

    solid_selection_star = parse_solid_selection(foo_pipeline, ['add_nums*'])
    assert len(solid_selection_star) == 3
    assert set(solid_selection_star) == {'add_nums', 'multiply_two', 'add_one'}

    solid_selection_both = parse_solid_selection(foo_pipeline, ['*add_nums+'])
    assert len(solid_selection_both) == 4
    assert set(solid_selection_both) == {'return_one', 'return_two', 'add_nums', 'multiply_two'}


def test_parse_solid_selection_multi():
    solid_selection_multi_disjoint = parse_solid_selection(
        foo_pipeline, ['return_one', 'add_nums+']
    )
    assert len(solid_selection_multi_disjoint) == 3
    assert set(solid_selection_multi_disjoint) == {'return_one', 'add_nums', 'multiply_two'}

    solid_selection_multi_overlap = parse_solid_selection(
        foo_pipeline, ['*add_nums', 'return_one+']
    )
    assert len(solid_selection_multi_overlap) == 3
    assert set(solid_selection_multi_overlap) == {'return_one', 'return_two', 'add_nums'}

    solid_selection_multi_with_invalid = parse_solid_selection(foo_pipeline, ['*add_nums', 'a'])
    assert len(solid_selection_multi_with_invalid) == 3
    assert set(solid_selection_multi_with_invalid) == {'return_one', 'return_two', 'add_nums'}


def test_parse_solid_selection_invalid():
    result = parse_solid_selection(foo_pipeline, ['1+some_solid'])
    assert result == set()

    result_comma = parse_solid_selection(foo_pipeline, ['some,solid'])
    assert result_comma == set()



from abc import ABCMeta, abstractmethod, abstractproperty

import six

from dagster import check
from dagster.core.execution.context.system import SystemExecutionContext
from dagster.core.execution.plan.objects import StepOutputHandle
from dagster.core.types.dagster_type import DagsterType, resolve_dagster_type

from .object_store import FilesystemObjectStore, InMemoryObjectStore, ObjectStore
from .type_storage import TypeStoragePluginRegistry


class IntermediateStorage(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def get_intermediate(self, context, dagster_type=None, step_output_handle=None):
        pass

    @abstractmethod
    def set_intermediate(self, context, dagster_type=None, step_output_handle=None, value=None):
        pass

    @abstractmethod
    def has_intermediate(self, context, step_output_handle):
        pass

    @abstractmethod
    def copy_intermediate_from_run(self, context, run_id, step_output_handle):
        pass

    @abstractproperty
    def is_persistent(self):
        pass

    def all_inputs_covered(self, context, step):
        return len(self.uncovered_inputs(context, step)) == 0

    def uncovered_inputs(self, context, step):
        from dagster.core.execution.plan.objects import ExecutionStep

        check.inst_param(step, 'step', ExecutionStep)
        uncovered_inputs = []
        for step_input in step.step_inputs:

            if step_input.is_from_single_output:
                for source_handle in step_input.source_handles:
                    if not self.has_intermediate(context, source_handle):
                        uncovered_inputs.append(source_handle)

            elif step_input.is_from_multiple_outputs:
                missing_source_handles = [
                    source_handle
                    for source_handle in step_input.source_handles
                    if not self.has_intermediate(context, source_handle)
                ]
                # only report as uncovered if all are missing from a multi-dep input
                if len(missing_source_handles) == len(step_input.source_handles):
                    uncovered_inputs = uncovered_inputs + missing_source_handles

        return uncovered_inputs


class InMemoryIntermediateStorage(IntermediateStorage):
    def __init__(self):

        self.values = {}

    # Note:
    # For the in-memory manager context and runtime are currently optional
    # because they are not strictly required. So we allow one to access
    # these values in places where those are not immediately available
    # but one wants to inspect intermediates. This is useful in test contexts
    # especially

    def get_intermediate(
        self, context, dagster_type=None, step_output_handle=None,
    ):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.opt_inst_param(dagster_type, 'dagster_type', DagsterType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        return self.values[step_output_handle]

    def set_intermediate(
        self, context, dagster_type=None, step_output_handle=None, value=None,
    ):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.opt_inst_param(dagster_type, 'dagster_type', DagsterType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        self.values[step_output_handle] = value

    def has_intermediate(self, context, step_output_handle):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        return step_output_handle in self.values

    def copy_intermediate_from_run(self, context, run_id, step_output_handle):
        check.failed('not implemented in in memory')

    @property
    def is_persistent(self):
        return False


class ObjectStoreIntermediateStorage(IntermediateStorage):
    def __init__(self, object_store, root_for_run_id, run_id, type_storage_plugin_registry):
        self.root_for_run_id = check.callable_param(root_for_run_id, 'root_for_run_id')
        self.run_id = check.str_param(run_id, 'run_id')
        self.object_store = check.inst_param(object_store, 'object_store', ObjectStore)
        self.type_storage_plugin_registry = check.inst_param(
            type_storage_plugin_registry, 'type_storage_plugin_registry', TypeStoragePluginRegistry
        )

    def _get_paths(self, step_output_handle):
        return ['intermediates', step_output_handle.step_key, step_output_handle.output_name]

    def get_intermediate_object(self, dagster_type, step_output_handle):
        check.inst_param(dagster_type, 'dagster_type', DagsterType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        paths = self._get_paths(step_output_handle)
        check.param_invariant(len(paths) > 0, 'paths')

        key = self.object_store.key_for_paths([self.root] + paths)
        return self.object_store.get_object(
            key, serialization_strategy=dagster_type.serialization_strategy
        )

    def get_intermediate(
        self, context, dagster_type=None, step_output_handle=None,
    ):
        dagster_type = resolve_dagster_type(dagster_type)
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.inst_param(dagster_type, 'dagster_type', DagsterType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        check.invariant(self.has_intermediate(context, step_output_handle))

        if self.type_storage_plugin_registry.is_registered(dagster_type):
            return self.type_storage_plugin_registry.get(dagster_type.name).get_intermediate_object(
                self, context, dagster_type, step_output_handle
            )
        elif dagster_type.name is None:
            self.type_storage_plugin_registry.check_for_unsupported_composite_overrides(
                dagster_type
            )

        return self.get_intermediate_object(dagster_type, step_output_handle)

    def set_intermediate_object(self, dagster_type, step_output_handle, value):
        check.inst_param(dagster_type, 'dagster_type', DagsterType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        paths = self._get_paths(step_output_handle)
        check.param_invariant(len(paths) > 0, 'paths')

        key = self.object_store.key_for_paths([self.root] + paths)
        return self.object_store.set_object(
            key, value, serialization_strategy=dagster_type.serialization_strategy
        )

    def set_intermediate(
        self, context, dagster_type=None, step_output_handle=None, value=None,
    ):
        dagster_type = resolve_dagster_type(dagster_type)
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.inst_param(dagster_type, 'dagster_type', DagsterType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)

        if self.has_intermediate(context, step_output_handle):
            context.log.warning(
                'Replacing existing intermediate for %s.%s'
                % (step_output_handle.step_key, step_output_handle.output_name)
            )

        if self.type_storage_plugin_registry.is_registered(dagster_type):
            return self.type_storage_plugin_registry.get(dagster_type.name).set_intermediate_object(
                self, context, dagster_type, step_output_handle, value
            )
        elif dagster_type.name is None:
            self.type_storage_plugin_registry.check_for_unsupported_composite_overrides(
                dagster_type
            )

        return self.set_intermediate_object(dagster_type, step_output_handle, value)

    def has_intermediate(self, context, step_output_handle):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        paths = self._get_paths(step_output_handle)
        check.list_param(paths, 'paths', of_type=str)
        check.param_invariant(len(paths) > 0, 'paths')

        key = self.object_store.key_for_paths([self.root] + paths)
        return self.object_store.has_object(key)

    def rm_intermediate(self, context, step_output_handle):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        paths = self._get_paths(step_output_handle)
        check.param_invariant(len(paths) > 0, 'paths')
        key = self.object_store.key_for_paths([self.root] + paths)
        return self.object_store.rm_object(key)

    def copy_intermediate_from_run(self, context, run_id, step_output_handle):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.str_param(run_id, 'run_id')
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        paths = self._get_paths(step_output_handle)

        src = self.object_store.key_for_paths([self.root_for_run_id(run_id)] + paths)
        dst = self.object_store.key_for_paths([self.root] + paths)

        return self.object_store.cp_object(src, dst)

    def uri_for_paths(self, paths, protocol=None):
        check.list_param(paths, 'paths', of_type=str)
        check.param_invariant(len(paths) > 0, 'paths')
        key = self.key_for_paths(paths)
        return self.object_store.uri_for_key(key, protocol)

    def key_for_paths(self, paths):
        return self.object_store.key_for_paths([self.root] + paths)

    @property
    def is_persistent(self):
        if isinstance(self.object_store, InMemoryObjectStore):
            return False
        return True

    @property
    def root(self):
        return self.root_for_run_id(self.run_id)


def build_in_mem_intermediates_storage(run_id, type_storage_plugin_registry=None):
    return ObjectStoreIntermediateStorage(
        InMemoryObjectStore(),
        lambda _: '',
        run_id,
        type_storage_plugin_registry
        if type_storage_plugin_registry
        else TypeStoragePluginRegistry(types_to_register=[]),
    )


def build_fs_intermediate_storage(root_for_run_id, run_id, type_storage_plugin_registry=None):
    return ObjectStoreIntermediateStorage(
        FilesystemObjectStore(),
        root_for_run_id,
        run_id,
        type_storage_plugin_registry
        if type_storage_plugin_registry
        else TypeStoragePluginRegistry(types_to_register=[]),
    )



'''Test that compute log tail processes go away when the parent hard crashes.'''

import sys

from dagster.core.execution.compute_logs import mirror_stream_to_file
from dagster.utils import segfault

if __name__ == '__main__':
    stdout_pids_file, stderr_pids_file = (sys.argv[1], sys.argv[2])
    with mirror_stream_to_file(sys.stdout, stdout_pids_file) as stdout_pids:
        with mirror_stream_to_file(sys.stderr, stderr_pids_file) as stderr_pids:
            sys.stdout.write('stdout pids: {pids}'.format(pids=str(stdout_pids)))
            sys.stdout.flush()
            sys.stderr.write('stderr pids: {pids}'.format(pids=str(stderr_pids)))
            sys.stderr.flush()
            segfault()



import os
import time
from collections import defaultdict

import pytest

from dagster import (
    DagsterEventType,
    Output,
    OutputDefinition,
    PipelineRun,
    RetryRequested,
    execute_pipeline,
    execute_pipeline_iterator,
    lambda_solid,
    pipeline,
    reconstructable,
    reexecute_pipeline,
    seven,
    solid,
)
from dagster.core.execution.api import create_execution_plan, execute_plan
from dagster.core.execution.retries import Retries, RetryMode
from dagster.core.instance import DagsterInstance

executors = pytest.mark.parametrize(
    "environment",
    [
        {'storage': {'filesystem': {}}},
        {'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
    ],
)


def define_run_retry_pipeline():
    @solid(config_schema={'fail': bool})
    def can_fail(context, _start_fail):
        if context.solid_config['fail']:
            raise Exception('blah')

        return 'okay perfect'

    @solid(
        output_defs=[
            OutputDefinition(bool, 'start_fail', is_required=False),
            OutputDefinition(bool, 'start_skip', is_required=False),
        ]
    )
    def two_outputs(_):
        yield Output(True, 'start_fail')
        # won't yield start_skip

    @solid
    def will_be_skipped(_, _start_skip):
        pass  # doesn't matter

    @solid
    def downstream_of_failed(_, input_str):
        return input_str

    @pipeline
    def pipe():
        start_fail, start_skip = two_outputs()
        downstream_of_failed(can_fail(start_fail))
        will_be_skipped(will_be_skipped(start_skip))

    return pipe


@executors
def test_retries(environment):
    instance = DagsterInstance.local_temp()
    pipe = reconstructable(define_run_retry_pipeline)
    fails = dict(environment)
    fails['solids'] = {'can_fail': {'config': {'fail': True}}}

    result = execute_pipeline(pipe, run_config=fails, instance=instance, raise_on_error=False,)

    assert not result.success

    passes = dict(environment)
    passes['solids'] = {'can_fail': {'config': {'fail': False}}}

    second_result = reexecute_pipeline(
        pipe, parent_run_id=result.run_id, run_config=passes, instance=instance,
    )
    assert second_result.success
    downstream_of_failed = second_result.result_for_solid('downstream_of_failed').output_value()
    assert downstream_of_failed == 'okay perfect'

    will_be_skipped = [
        e for e in second_result.event_list if 'will_be_skipped' in str(e.solid_handle)
    ]
    assert str(will_be_skipped[0].event_type_value) == 'STEP_SKIPPED'
    assert str(will_be_skipped[1].event_type_value) == 'STEP_SKIPPED'


def define_step_retry_pipeline():
    @solid(config_schema=str)
    def fail_first_time(context):
        file = os.path.join(context.solid_config, 'i_threw_up')
        if os.path.exists(file):
            return 'okay perfect'
        else:
            open(file, 'a').close()
            raise RetryRequested()

    @pipeline
    def step_retry():
        fail_first_time()

    return step_retry


@executors
def test_step_retry(environment):
    with seven.TemporaryDirectory() as tempdir:
        env = dict(environment)
        env['solids'] = {'fail_first_time': {'config': tempdir}}
        result = execute_pipeline(
            reconstructable(define_step_retry_pipeline),
            run_config=env,
            instance=DagsterInstance.local_temp(),
        )
    assert result.success
    events = defaultdict(list)
    for ev in result.event_list:
        events[ev.event_type].append(ev)

    assert len(events[DagsterEventType.STEP_START]) == 1
    assert len(events[DagsterEventType.STEP_UP_FOR_RETRY]) == 1
    assert len(events[DagsterEventType.STEP_RESTARTED]) == 1
    assert len(events[DagsterEventType.STEP_SUCCESS]) == 1


def define_retry_limit_pipeline():
    @lambda_solid
    def default_max():
        raise RetryRequested()

    @lambda_solid
    def three_max():
        raise RetryRequested(max_retries=3)

    @pipeline
    def retry_limits():
        default_max()
        three_max()

    return retry_limits


@executors
def test_step_retry_limit(environment):
    result = execute_pipeline(
        reconstructable(define_retry_limit_pipeline),
        run_config=environment,
        raise_on_error=False,
        instance=DagsterInstance.local_temp(),
    )
    assert not result.success

    events = defaultdict(list)
    for ev in result.events_by_step_key['default_max.compute']:
        events[ev.event_type].append(ev)

    assert len(events[DagsterEventType.STEP_START]) == 1
    assert len(events[DagsterEventType.STEP_UP_FOR_RETRY]) == 1
    assert len(events[DagsterEventType.STEP_RESTARTED]) == 1
    assert len(events[DagsterEventType.STEP_FAILURE]) == 1

    events = defaultdict(list)
    for ev in result.events_by_step_key['three_max.compute']:
        events[ev.event_type].append(ev)

    assert len(events[DagsterEventType.STEP_START]) == 1
    assert len(events[DagsterEventType.STEP_UP_FOR_RETRY]) == 3
    assert len(events[DagsterEventType.STEP_RESTARTED]) == 3
    assert len(events[DagsterEventType.STEP_FAILURE]) == 1


def test_retry_deferral():
    events = execute_plan(
        create_execution_plan(define_retry_limit_pipeline()),
        pipeline_run=PipelineRun(pipeline_name='retry_limits', run_id='42'),
        retries=Retries(RetryMode.DEFERRED),
        instance=DagsterInstance.local_temp(),
    )
    events_by_type = defaultdict(list)
    for ev in events:
        events_by_type[ev.event_type].append(ev)

    assert len(events_by_type[DagsterEventType.STEP_START]) == 2
    assert len(events_by_type[DagsterEventType.STEP_UP_FOR_RETRY]) == 2
    assert DagsterEventType.STEP_RESTARTED not in events
    assert DagsterEventType.STEP_SUCCESS not in events


DELAY = 2


def define_retry_wait_fixed_pipeline():
    @solid(config_schema=str)
    def fail_first_and_wait(context):
        file = os.path.join(context.solid_config, 'i_threw_up')
        if os.path.exists(file):
            return 'okay perfect'
        else:
            open(file, 'a').close()
            raise RetryRequested(seconds_to_wait=DELAY)

    @pipeline
    def step_retry():
        fail_first_and_wait()

    return step_retry


@executors
def test_step_retry_fixed_wait(environment):
    with seven.TemporaryDirectory() as tempdir:
        env = dict(environment)
        env['solids'] = {'fail_first_and_wait': {'config': tempdir}}

        event_iter = execute_pipeline_iterator(
            reconstructable(define_retry_wait_fixed_pipeline),
            run_config=env,
            instance=DagsterInstance.local_temp(),
        )
        start_wait = None
        end_wait = None
        success = None
        for event in event_iter:
            if event.is_step_up_for_retry:
                start_wait = time.time()
            if event.is_step_restarted:
                end_wait = time.time()
            if event.is_pipeline_success:
                success = True

        assert success
        assert start_wait is not None
        assert end_wait is not None
        delay = end_wait - start_wait
        assert delay > DELAY



from dagster import check
from dagster.core.errors import DagsterUserCodeProcessError
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster.utils.error import SerializableErrorInfo

from .utils import execute_unary_api_cli_command


def sync_list_repositories(executable_path, python_file, module_name, working_directory, attribute):
    from dagster.grpc.types import ListRepositoriesResponse, ListRepositoriesInput

    result = check.inst(
        execute_unary_api_cli_command(
            executable_path,
            'list_repositories',
            ListRepositoriesInput(
                module_name=module_name,
                python_file=python_file,
                working_directory=working_directory,
                attribute=attribute,
            ),
        ),
        (ListRepositoriesResponse, SerializableErrorInfo),
    )
    if isinstance(result, SerializableErrorInfo):
        raise DagsterUserCodeProcessError(
            result.to_string(), user_code_process_error_infos=[result]
        )
    else:
        return result


def sync_list_repositories_grpc(api_client):
    from dagster.grpc.client import DagsterGrpcClient
    from dagster.grpc.types import ListRepositoriesResponse

    check.inst_param(api_client, 'api_client', DagsterGrpcClient)
    result = check.inst(
        api_client.list_repositories(), (ListRepositoriesResponse, SerializableErrorInfo)
    )
    if isinstance(result, SerializableErrorInfo):
        raise DagsterUserCodeProcessError(
            result.to_string(), user_code_process_error_infos=[result]
        )
    else:
        return result


def sync_list_repositories_ephemeral_grpc(
    executable_path, python_file, module_name, working_directory, attribute
):
    from dagster.grpc.client import ephemeral_grpc_api_client

    check.str_param(executable_path, 'executable_path')
    check.opt_str_param(python_file, 'python_file')
    check.opt_str_param(module_name, 'module_name')
    check.opt_str_param(working_directory, 'working_directory')

    with ephemeral_grpc_api_client(
        loadable_target_origin=LoadableTargetOrigin(
            executable_path=executable_path,
            module_name=module_name,
            python_file=python_file,
            working_directory=working_directory,
            attribute=attribute,
        )
    ) as api_client:
        return sync_list_repositories_grpc(api_client)



from .query_parser import (
    NameToken,
    PlusToken,
    SelectionParserException,
    WhitespaceToken,
    parse_selector_text,
    tokenize_selector_text,
)
from .subset_selector import (
    MAX_NUM,
    Traverser,
    generate_dep_graph,
    parse_clause,
    parse_solid_selection,
)



import pytest

from dagster import (
    ModeDefinition,
    check,
    configured,
    execute_pipeline,
    intermediate_storage,
    pipeline,
    system_storage,
)
from dagster.core.storage.object_store import InMemoryObjectStore
from dagster.core.storage.system_storage import (
    build_intermediate_storage_from_object_store,
    create_mem_system_storage_data,
)


def create_mem_system_intermediate_store(init_context):
    object_store = InMemoryObjectStore()
    return build_intermediate_storage_from_object_store(
        object_store=object_store, init_context=init_context
    )


def assert_pipeline_runs_with_storage(storage_def, storage_config):
    @pipeline(mode_defs=[ModeDefinition(system_storage_defs=[storage_def])])
    def pass_pipeline():
        pass

    result = execute_pipeline(pass_pipeline, {'storage': storage_config})
    assert result.success


def assert_pipeline_runs_with_intermediate_storage(
    intermediate_storage_def, intermediate_storage_config
):
    @pipeline(mode_defs=[ModeDefinition(intermediate_storage_defs=[intermediate_storage_def])])
    def pass_pipeline():
        pass

    result = execute_pipeline(pass_pipeline, {'intermediate_storage': intermediate_storage_config})
    assert result.success


def test_naked_system_storage():
    it = {}

    @system_storage(required_resource_keys=set())
    def no_config_storage(init_context):
        it['ran'] = True
        return create_mem_system_storage_data(init_context)

    assert_pipeline_runs_with_storage(no_config_storage, {'no_config_storage': {}})
    assert it['ran']

    it = {}
    assert_pipeline_runs_with_storage(no_config_storage, {'no_config_storage': None})
    assert it['ran']


@pytest.mark.xfail(raises=check.ParameterCheckError)
def test_system_storage_primitive_config():
    it = {}

    @system_storage(required_resource_keys=set(), config_schema=str)
    def test_storage(init_context):
        assert init_context.system_storage_config == 'secret testing value!!'
        it['ran'] = True
        return create_mem_system_storage_data(init_context)

    assert_pipeline_runs_with_storage(
        test_storage, {'test_storage': {'config': 'secret testing value!!'}}
    )
    assert it['ran']


def test_system_storage_dict_config():
    it = {}

    @system_storage(required_resource_keys=set(), config_schema={'value': str})
    def test_storage(init_context):
        assert init_context.system_storage_config['value'] == 'secret testing value!!'
        it['ran'] = True
        return create_mem_system_storage_data(init_context)

    assert_pipeline_runs_with_storage(
        test_storage, {'test_storage': {'config': {'value': 'secret testing value!!'}}}
    )
    assert it['ran']


def test_system_storage_dict_config_configured():
    it = {}

    @system_storage(required_resource_keys=set(), config_schema={'value': str})
    def test_storage(init_context):
        assert init_context.system_storage_config['value'] == 'secret testing value!!'
        it['ran'] = True
        return create_mem_system_storage_data(init_context)

    test_storage_configured = configured(test_storage)({'value': 'secret testing value!!'})

    assert_pipeline_runs_with_storage(test_storage_configured, {'test_storage': {}})
    assert it['ran']

    it = {}
    assert_pipeline_runs_with_storage(test_storage_configured, {'test_storage': None})
    assert it['ran']


def test_naked_intermediate_storage():
    it = {}

    @intermediate_storage(required_resource_keys=set())
    def no_config_intermediate_storage(init_context):
        it['ran'] = True
        return create_mem_system_intermediate_store(init_context)

    assert_pipeline_runs_with_intermediate_storage(
        no_config_intermediate_storage, {'no_config_intermediate_storage': {}}
    )
    assert it['ran']

    it = {}
    assert_pipeline_runs_with_intermediate_storage(
        no_config_intermediate_storage, {'no_config_intermediate_storage': None}
    )
    assert it['ran']


@pytest.mark.xfail(raises=check.ParameterCheckError)
def test_intermediate_storage_primitive_config():
    it = {}

    @intermediate_storage(required_resource_keys=set(), config_schema=str)
    def test_intermediate_storage(init_context):
        assert init_context.intermediate_storage_config == 'secret testing value!!'
        it['ran'] = True
        return create_mem_system_intermediate_store(init_context)

    assert_pipeline_runs_with_intermediate_storage(
        test_intermediate_storage,
        {'test_intermediate_storage': {'config': 'secret testing value!!'}},
    )
    assert it['ran']


def test_intermediate_storage_dict_config():
    it = {}

    @intermediate_storage(required_resource_keys=set(), config_schema={'value': str})
    def test_intermediate_storage(init_context):
        assert init_context.intermediate_storage_config['value'] == 'secret testing value!!'
        it['ran'] = True
        return create_mem_system_intermediate_store(init_context)

    assert_pipeline_runs_with_intermediate_storage(
        test_intermediate_storage,
        {'test_intermediate_storage': {'config': {'value': 'secret testing value!!'}}},
    )
    assert it['ran']


def test_intermediate_storage_dict_config_configured():
    it = {}

    @intermediate_storage(required_resource_keys=set(), config_schema={'value': str})
    def test_intermediate_storage(init_context):
        assert init_context.intermediate_storage_config['value'] == 'secret testing value!!'
        it['ran'] = True
        return create_mem_system_intermediate_store(init_context)

    test_intermediate_storage_configured = configured(test_intermediate_storage)(
        {'value': 'secret testing value!!'}
    )

    assert_pipeline_runs_with_intermediate_storage(
        test_intermediate_storage_configured, {'test_intermediate_storage': {}}
    )
    assert it['ran']

    it = {}
    assert_pipeline_runs_with_intermediate_storage(
        test_intermediate_storage_configured, {'test_intermediate_storage': None}
    )
    assert it['ran']



from __future__ import print_function

import re

import pytest
from click.testing import CliRunner

from dagster.cli.pipeline import execute_scaffold_command, pipeline_scaffold_command

from .test_cli_commands import (
    valid_external_pipeline_target_args,
    valid_external_pipeline_target_cli_args_no_preset,
)


def no_print(_):
    return None


@pytest.mark.parametrize('scaffold_args', valid_external_pipeline_target_args())
def test_scaffold_command(scaffold_args):
    cli_args, uses_legacy_repository_yaml_format = scaffold_args
    if uses_legacy_repository_yaml_format:
        with pytest.warns(
            UserWarning,
            match=re.escape(
                'You are using the legacy repository yaml format. Please update your file '
            ),
        ):
            cli_args['print_only_required'] = True
            execute_scaffold_command(cli_args=cli_args, print_fn=no_print)

            cli_args['print_only_required'] = False
            execute_scaffold_command(cli_args=cli_args, print_fn=no_print)
    else:
        cli_args['print_only_required'] = True
        execute_scaffold_command(cli_args=cli_args, print_fn=no_print)

        cli_args['print_only_required'] = False
        execute_scaffold_command(cli_args=cli_args, print_fn=no_print)


@pytest.mark.parametrize('execute_cli_args', valid_external_pipeline_target_cli_args_no_preset())
def test_scaffold_command_cli(execute_cli_args):
    cli_args, uses_legacy_repository_yaml_format = execute_cli_args

    runner = CliRunner()

    if uses_legacy_repository_yaml_format:
        with pytest.warns(
            UserWarning,
            match=re.escape(
                'You are using the legacy repository yaml format. Please update your file '
            ),
        ):
            result = runner.invoke(pipeline_scaffold_command, cli_args)
            assert result.exit_code == 0

            result = runner.invoke(pipeline_scaffold_command, ['--print-only-required'] + cli_args)
            assert result.exit_code == 0
    else:
        result = runner.invoke(pipeline_scaffold_command, cli_args)
        assert result.exit_code == 0

        result = runner.invoke(pipeline_scaffold_command, ['--print-only-required'] + cli_args)
        assert result.exit_code == 0



"""add snapshots to run storage

Revision ID: c63a27054f08
Revises: 9fe9e746268c
Create Date: 2020-04-09 05:57:20.639458

"""
import sqlalchemy as sa
from alembic import op

from dagster.core.storage.migration.utils import has_column, has_table

# alembic magic breaks pylint
# pylint: disable=no-member

# revision identifiers, used by Alembic.
revision = 'c63a27054f08'
down_revision = '9fe9e746268c'
branch_labels = None
depends_on = None


def upgrade():
    if not has_table('snapshots'):
        op.create_table(
            'snapshots',
            sa.Column('id', sa.Integer, primary_key=True, autoincrement=True, nullable=False),
            sa.Column('snapshot_id', sa.String(255), unique=True, nullable=False),
            sa.Column('snapshot_body', sa.LargeBinary, nullable=False),
            sa.Column('snapshot_type', sa.String(63), nullable=False),
        )

    if not has_column('runs', 'snapshot_id'):
        # Sqlite does not support adding foreign keys to existing
        # tables, so we are forced to fallback on this witchcraft.
        # See https://alembic.sqlalchemy.org/en/latest/batch.html#dealing-with-referencing-foreign-keys
        # for additional context
        with op.batch_alter_table('runs') as batch_op:
            batch_op.execute('PRAGMA foreign_keys = OFF;')
            batch_op.add_column(
                sa.Column(
                    'snapshot_id',
                    sa.String(255),
                    sa.ForeignKey(
                        'snapshots.snapshot_id', name='fk_runs_snapshot_id_snapshots_snapshot_id'
                    ),
                ),
            )
        op.execute('PRAGMA foreign_keys = ON;')


def downgrade():
    if has_column('runs', 'snapshot_id'):
        with op.batch_alter_table('runs') as batch_op:
            batch_op.drop_column('snapshot_id')

    if has_table('snapshots'):
        op.drop_table('snapshots')



# Utility script for managing compute log stdout mirroring
#
# This script is run as a subprocess during step computation to babysit a tail child process which
# mirrors compute log output from a file to stdout.  This utility script checks to see if its parent
# process has died and kills the tail child process if so.  This is to ensure that execution that
# suddenly exits mid-computation without cleaning up after itself will not orphan long-lived tail
# processes.

import os
import signal
import sys
import time


def watch(args):
    if not args or len(args) != 2:
        return

    parent_pid = int(args[0])
    tail_pid = int(args[1])

    if not parent_pid or not tail_pid:
        return

    while True:
        # check if this process has been orphaned, in which case kill the tail_pid
        if os.getppid() == 1:
            try:
                os.kill(tail_pid, signal.SIGTERM)
            except OSError:
                pass
            break
        else:
            time.sleep(1)


if __name__ == '__main__':
    watch(sys.argv[1:])



from dagster import EventMetadataEntry, Failure, execute_pipeline, lambda_solid, pipeline


def test_failure():
    @lambda_solid
    def throw():
        raise Failure(
            description='it Failure',
            metadata_entries=[
                EventMetadataEntry.text(label='label', text='text', description='description')
            ],
        )

    @pipeline
    def failure():
        throw()

    result = execute_pipeline(failure, raise_on_error=False)
    assert not result.success
    failure_data = result.result_for_solid('throw').failure_data
    assert failure_data
    assert failure_data.error.cls_name == 'Failure'

    # hard coded
    assert failure_data.user_failure_data.label == 'intentional-failure'
    # from Failure
    assert failure_data.user_failure_data.description == 'it Failure'
    assert failure_data.user_failure_data.metadata_entries[0].label == 'label'
    assert failure_data.user_failure_data.metadata_entries[0].entry_data.text == 'text'
    assert failure_data.user_failure_data.metadata_entries[0].description == 'description'



import pytest

from dagster.core.definitions import PipelineDefinition
from dagster.core.errors import DagsterRunAlreadyExists, DagsterSnapshotDoesNotExist
from dagster.core.snap import create_pipeline_snapshot_id
from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus, PipelineRunsFilter
from dagster.core.storage.tags import PARENT_RUN_ID_TAG, ROOT_RUN_ID_TAG
from dagster.core.utils import make_new_run_id
from dagster.serdes import serialize_pp


class TestRunStorage:
    '''
    You can extend this class to easily run these set of tests on any run storage. When extending,
    you simply need to override the `run_storage` fixture and return your implementation of
    `RunStorage`.

    For example:

    ```
    TestRunStorage.__test__ = False

    class TestMyStorageImplementation(TestRunStorage):
        __test__ = True

        @pytest.fixture(scope='function', name='storage')
        def run_storage(self):  # pylint: disable=arguments-differ
            return MyStorageImplementation()
    ```
    '''

    @pytest.fixture(name='storage', params=[])
    def run_storage(self, request):
        with request.param() as s:
            yield s

    @staticmethod
    def build_run(
        run_id,
        pipeline_name,
        mode='default',
        tags=None,
        status=PipelineRunStatus.NOT_STARTED,
        parent_run_id=None,
        root_run_id=None,
    ):
        return PipelineRun(
            pipeline_name=pipeline_name,
            run_id=run_id,
            run_config=None,
            mode=mode,
            tags=tags,
            status=status,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
        )

    def test_basic_storage(self, storage):
        assert storage
        run_id = make_new_run_id()
        added = storage.add_run(
            TestRunStorage.build_run(
                run_id=run_id, pipeline_name='some_pipeline', tags={'foo': 'bar'}
            )
        )
        assert added
        runs = storage.get_runs()
        assert len(runs) == 1
        run = runs[0]
        assert run.run_id == run_id
        assert run.pipeline_name == 'some_pipeline'
        assert run.tags
        assert run.tags.get('foo') == 'bar'
        assert storage.has_run(run_id)
        fetched_run = storage.get_run_by_id(run_id)
        assert fetched_run.run_id == run_id
        assert fetched_run.pipeline_name == 'some_pipeline'

    def test_clear(self, storage):
        assert storage
        run_id = make_new_run_id()
        storage.add_run(TestRunStorage.build_run(run_id=run_id, pipeline_name='some_pipeline'))
        assert len(storage.get_runs()) == 1
        storage.wipe()
        assert list(storage.get_runs()) == []

    def test_fetch_by_pipeline(self, storage):
        assert storage
        one = make_new_run_id()
        two = make_new_run_id()
        storage.add_run(TestRunStorage.build_run(run_id=one, pipeline_name='some_pipeline'))
        storage.add_run(TestRunStorage.build_run(run_id=two, pipeline_name='some_other_pipeline'))
        assert len(storage.get_runs()) == 2
        some_runs = storage.get_runs(PipelineRunsFilter(pipeline_name='some_pipeline'))
        assert len(some_runs) == 1
        assert some_runs[0].run_id == one

    def test_fetch_by_filter(self, storage):
        assert storage
        one = make_new_run_id()
        two = make_new_run_id()
        three = make_new_run_id()

        storage.add_run(
            TestRunStorage.build_run(
                run_id=one,
                pipeline_name='some_pipeline',
                tags={'tag': 'hello', 'tag2': 'world'},
                status=PipelineRunStatus.SUCCESS,
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=two,
                pipeline_name='some_pipeline',
                tags={'tag': 'hello'},
                status=PipelineRunStatus.FAILURE,
            ),
        )

        storage.add_run(
            TestRunStorage.build_run(
                run_id=three, pipeline_name='other_pipeline', status=PipelineRunStatus.SUCCESS
            )
        )

        assert len(storage.get_runs()) == 3

        some_runs = storage.get_runs(PipelineRunsFilter(run_ids=[one]))
        count = storage.get_runs_count(PipelineRunsFilter(run_ids=[one]))
        assert len(some_runs) == 1
        assert count == 1
        assert some_runs[0].run_id == one

        some_runs = storage.get_runs(PipelineRunsFilter(pipeline_name='some_pipeline'))
        count = storage.get_runs_count(PipelineRunsFilter(pipeline_name='some_pipeline'))
        assert len(some_runs) == 2
        assert count == 2
        assert some_runs[0].run_id == two
        assert some_runs[1].run_id == one

        some_runs = storage.get_runs(PipelineRunsFilter(status=PipelineRunStatus.SUCCESS))
        count = storage.get_runs_count(PipelineRunsFilter(status=PipelineRunStatus.SUCCESS))
        assert len(some_runs) == 2
        assert count == 2
        assert some_runs[0].run_id == three
        assert some_runs[1].run_id == one

        some_runs = storage.get_runs(PipelineRunsFilter(tags={'tag': 'hello'}))
        count = storage.get_runs_count(PipelineRunsFilter(tags={'tag': 'hello'}))
        assert len(some_runs) == 2
        assert count == 2
        assert some_runs[0].run_id == two
        assert some_runs[1].run_id == one

        some_runs = storage.get_runs(PipelineRunsFilter(tags={'tag': 'hello', 'tag2': 'world'}))
        count = storage.get_runs_count(PipelineRunsFilter(tags={'tag': 'hello', 'tag2': 'world'}))
        assert len(some_runs) == 1
        assert count == 1
        assert some_runs[0].run_id == one

        some_runs = storage.get_runs(
            PipelineRunsFilter(pipeline_name="some_pipeline", tags={'tag': 'hello'})
        )
        count = storage.get_runs_count(
            PipelineRunsFilter(pipeline_name="some_pipeline", tags={'tag': 'hello'})
        )
        assert len(some_runs) == 2
        assert count == 2
        assert some_runs[0].run_id == two
        assert some_runs[1].run_id == one

        some_runs = storage.get_runs(
            PipelineRunsFilter(
                pipeline_name="some_pipeline",
                tags={'tag': 'hello'},
                status=PipelineRunStatus.SUCCESS,
            )
        )
        count = storage.get_runs_count(
            PipelineRunsFilter(
                pipeline_name="some_pipeline",
                tags={'tag': 'hello'},
                status=PipelineRunStatus.SUCCESS,
            )
        )
        assert len(some_runs) == 1
        assert count == 1
        assert some_runs[0].run_id == one

        # All filters
        some_runs = storage.get_runs(
            PipelineRunsFilter(
                run_ids=[one],
                pipeline_name="some_pipeline",
                tags={'tag': 'hello'},
                status=PipelineRunStatus.SUCCESS,
            )
        )
        count = storage.get_runs_count(
            PipelineRunsFilter(
                run_ids=[one],
                pipeline_name="some_pipeline",
                tags={'tag': 'hello'},
                status=PipelineRunStatus.SUCCESS,
            )
        )
        assert len(some_runs) == 1
        assert count == 1
        assert some_runs[0].run_id == one

        some_runs = storage.get_runs(PipelineRunsFilter())
        count = storage.get_runs_count(PipelineRunsFilter())
        assert len(some_runs) == 3
        assert count == 3

    def test_fetch_count_by_tag(self, storage):
        assert storage
        one = make_new_run_id()
        two = make_new_run_id()
        three = make_new_run_id()
        storage.add_run(
            TestRunStorage.build_run(
                run_id=one,
                pipeline_name='some_pipeline',
                tags={'mytag': 'hello', 'mytag2': 'world'},
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=two,
                pipeline_name='some_pipeline',
                tags={'mytag': 'goodbye', 'mytag2': 'world'},
            )
        )
        storage.add_run(TestRunStorage.build_run(run_id=three, pipeline_name='some_pipeline'))
        assert len(storage.get_runs()) == 3

        run_count = storage.get_runs_count(
            filters=PipelineRunsFilter(tags={'mytag': 'hello', 'mytag2': 'world'})
        )
        assert run_count == 1

        run_count = storage.get_runs_count(filters=PipelineRunsFilter(tags={'mytag2': 'world'}))
        assert run_count == 2

        run_count = storage.get_runs_count()
        assert run_count == 3

        assert storage.get_run_tags() == [('mytag', {'hello', 'goodbye'}), ('mytag2', {'world'})]

    def test_fetch_by_tags(self, storage):
        assert storage
        one = make_new_run_id()
        two = make_new_run_id()
        three = make_new_run_id()
        storage.add_run(
            TestRunStorage.build_run(
                run_id=one,
                pipeline_name='some_pipeline',
                tags={'mytag': 'hello', 'mytag2': 'world'},
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=two,
                pipeline_name='some_pipeline',
                tags={'mytag': 'goodbye', 'mytag2': 'world'},
            )
        )
        storage.add_run(TestRunStorage.build_run(run_id=three, pipeline_name='some_pipeline'))
        assert len(storage.get_runs()) == 3

        some_runs = storage.get_runs(PipelineRunsFilter(tags={'mytag': 'hello', 'mytag2': 'world'}))

        assert len(some_runs) == 1
        assert some_runs[0].run_id == one

        some_runs = storage.get_runs(PipelineRunsFilter(tags={'mytag2': 'world'}))
        assert len(some_runs) == 2
        assert some_runs[0].run_id == two
        assert some_runs[1].run_id == one

        some_runs = storage.get_runs(PipelineRunsFilter(tags={}))
        assert len(some_runs) == 3

    def test_paginated_fetch(self, storage):
        assert storage
        one, two, three = [make_new_run_id(), make_new_run_id(), make_new_run_id()]
        storage.add_run(
            TestRunStorage.build_run(
                run_id=one, pipeline_name='some_pipeline', tags={'mytag': 'hello'}
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=two, pipeline_name='some_pipeline', tags={'mytag': 'hello'}
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=three, pipeline_name='some_pipeline', tags={'mytag': 'hello'}
            )
        )

        all_runs = storage.get_runs()
        assert len(all_runs) == 3
        sliced_runs = storage.get_runs(cursor=three, limit=1)
        assert len(sliced_runs) == 1
        assert sliced_runs[0].run_id == two

        all_runs = storage.get_runs(PipelineRunsFilter(pipeline_name='some_pipeline'))
        assert len(all_runs) == 3
        sliced_runs = storage.get_runs(
            PipelineRunsFilter(pipeline_name='some_pipeline'), cursor=three, limit=1
        )
        assert len(sliced_runs) == 1
        assert sliced_runs[0].run_id == two

        all_runs = storage.get_runs(PipelineRunsFilter(tags={'mytag': 'hello'}))
        assert len(all_runs) == 3
        sliced_runs = storage.get_runs(
            PipelineRunsFilter(tags={'mytag': 'hello'}), cursor=three, limit=1
        )
        assert len(sliced_runs) == 1
        assert sliced_runs[0].run_id == two

    def test_fetch_by_status(self, storage):
        assert storage
        one = make_new_run_id()
        two = make_new_run_id()
        three = make_new_run_id()
        four = make_new_run_id()
        storage.add_run(
            TestRunStorage.build_run(
                run_id=one, pipeline_name='some_pipeline', status=PipelineRunStatus.NOT_STARTED
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=two, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=three, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=four, pipeline_name='some_pipeline', status=PipelineRunStatus.FAILURE
            )
        )

        assert {
            run.run_id
            for run in storage.get_runs(PipelineRunsFilter(status=PipelineRunStatus.NOT_STARTED))
        } == {one}

        assert {
            run.run_id
            for run in storage.get_runs(PipelineRunsFilter(status=PipelineRunStatus.STARTED))
        } == {two, three,}

        assert {
            run.run_id
            for run in storage.get_runs(PipelineRunsFilter(status=PipelineRunStatus.FAILURE))
        } == {four}

        assert {
            run.run_id
            for run in storage.get_runs(PipelineRunsFilter(status=PipelineRunStatus.SUCCESS))
        } == set()

    def test_fetch_by_status_cursored(self, storage):
        assert storage
        one = make_new_run_id()
        two = make_new_run_id()
        three = make_new_run_id()
        four = make_new_run_id()
        storage.add_run(
            TestRunStorage.build_run(
                run_id=one, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=two, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=three, pipeline_name='some_pipeline', status=PipelineRunStatus.NOT_STARTED
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=four, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED
            )
        )

        cursor_four_runs = storage.get_runs(
            PipelineRunsFilter(status=PipelineRunStatus.STARTED), cursor=four
        )
        assert len(cursor_four_runs) == 2
        assert {run.run_id for run in cursor_four_runs} == {one, two}

        cursor_two_runs = storage.get_runs(
            PipelineRunsFilter(status=PipelineRunStatus.STARTED), cursor=two
        )
        assert len(cursor_two_runs) == 1
        assert {run.run_id for run in cursor_two_runs} == {one}

        cursor_one_runs = storage.get_runs(
            PipelineRunsFilter(status=PipelineRunStatus.STARTED), cursor=one
        )
        assert not cursor_one_runs

        cursor_four_limit_one = storage.get_runs(
            PipelineRunsFilter(status=PipelineRunStatus.STARTED), cursor=four, limit=1
        )
        assert len(cursor_four_limit_one) == 1
        assert cursor_four_limit_one[0].run_id == two

    def test_delete(self, storage):
        assert storage
        run_id = make_new_run_id()
        storage.add_run(TestRunStorage.build_run(run_id=run_id, pipeline_name='some_pipeline'))
        assert len(storage.get_runs()) == 1
        storage.delete_run(run_id)
        assert list(storage.get_runs()) == []

    def test_delete_with_tags(self, storage):
        assert storage
        run_id = make_new_run_id()
        storage.add_run(
            TestRunStorage.build_run(
                run_id=run_id, pipeline_name='some_pipeline', tags={run_id: run_id},
            )
        )
        assert len(storage.get_runs()) == 1
        assert run_id in [key for key, value in storage.get_run_tags()]
        storage.delete_run(run_id)
        assert list(storage.get_runs()) == []
        assert run_id not in [key for key, value in storage.get_run_tags()]

    def test_wipe_tags(self, storage):
        run_id = 'some_run_id'
        run = PipelineRun(run_id=run_id, pipeline_name='a_pipeline', tags={'foo': 'bar'})

        storage.add_run(run)

        assert storage.get_run_by_id(run_id) == run
        assert dict(storage.get_run_tags()) == {'foo': {'bar'}}

        storage.wipe()
        assert list(storage.get_runs()) == []
        assert dict(storage.get_run_tags()) == {}

    def test_write_conflicting_run_id(self, storage):
        double_run_id = 'double_run_id'
        pipeline_def = PipelineDefinition(name='some_pipeline', solid_defs=[])

        run = PipelineRun(run_id=double_run_id, pipeline_name=pipeline_def.name)

        assert storage.add_run(run)
        with pytest.raises(DagsterRunAlreadyExists):
            storage.add_run(run)

    def test_add_get_snapshot(self, storage):
        pipeline_def = PipelineDefinition(name='some_pipeline', solid_defs=[])
        pipeline_snapshot = pipeline_def.get_pipeline_snapshot()
        pipeline_snapshot_id = create_pipeline_snapshot_id(pipeline_snapshot)

        assert storage.add_pipeline_snapshot(pipeline_snapshot) == pipeline_snapshot_id
        fetched_pipeline_snapshot = storage.get_pipeline_snapshot(pipeline_snapshot_id)
        assert fetched_pipeline_snapshot
        assert serialize_pp(fetched_pipeline_snapshot) == serialize_pp(pipeline_snapshot)
        assert storage.has_pipeline_snapshot(pipeline_snapshot_id)
        assert not storage.has_pipeline_snapshot('nope')

        storage.wipe()

        assert not storage.has_pipeline_snapshot(pipeline_snapshot_id)

    def test_single_write_read_with_snapshot(self, storage):
        run_with_snapshot_id = 'lkasjdflkjasdf'
        pipeline_def = PipelineDefinition(name='some_pipeline', solid_defs=[])

        pipeline_snapshot = pipeline_def.get_pipeline_snapshot()

        pipeline_snapshot_id = create_pipeline_snapshot_id(pipeline_snapshot)

        run_with_snapshot = PipelineRun(
            run_id=run_with_snapshot_id,
            pipeline_name=pipeline_def.name,
            pipeline_snapshot_id=pipeline_snapshot_id,
        )

        assert not storage.has_pipeline_snapshot(pipeline_snapshot_id)

        assert storage.add_pipeline_snapshot(pipeline_snapshot) == pipeline_snapshot_id

        assert serialize_pp(storage.get_pipeline_snapshot(pipeline_snapshot_id)) == serialize_pp(
            pipeline_snapshot
        )

        storage.add_run(run_with_snapshot)

        assert storage.get_run_by_id(run_with_snapshot_id) == run_with_snapshot

        storage.wipe()

        assert not storage.has_pipeline_snapshot(pipeline_snapshot_id)
        assert not storage.has_run(run_with_snapshot_id)

    def test_single_write_with_missing_snapshot(self, storage):

        run_with_snapshot_id = 'lkasjdflkjasdf'
        pipeline_def = PipelineDefinition(name='some_pipeline', solid_defs=[])

        run_with_missing_snapshot = PipelineRun(
            run_id=run_with_snapshot_id,
            pipeline_name=pipeline_def.name,
            pipeline_snapshot_id='nope',
        )

        with pytest.raises(DagsterSnapshotDoesNotExist):
            storage.add_run(run_with_missing_snapshot)

    def test_add_get_execution_snapshot(self, storage):
        from dagster.core.execution.api import create_execution_plan
        from dagster.core.snap import snapshot_from_execution_plan

        pipeline_def = PipelineDefinition(name='some_pipeline', solid_defs=[])
        execution_plan = create_execution_plan(pipeline_def)
        ep_snapshot = snapshot_from_execution_plan(
            execution_plan, pipeline_def.get_pipeline_snapshot_id()
        )

        snapshot_id = storage.add_execution_plan_snapshot(ep_snapshot)
        fetched_ep_snapshot = storage.get_execution_plan_snapshot(snapshot_id)
        assert fetched_ep_snapshot
        assert serialize_pp(fetched_ep_snapshot) == serialize_pp(ep_snapshot)
        assert storage.has_execution_plan_snapshot(snapshot_id)
        assert not storage.has_execution_plan_snapshot('nope')

        storage.wipe()

        assert not storage.has_execution_plan_snapshot(snapshot_id)

    def test_fetch_run_filter(self, storage):
        assert storage
        one = make_new_run_id()
        two = make_new_run_id()

        storage.add_run(
            TestRunStorage.build_run(
                run_id=one, pipeline_name='some_pipeline', status=PipelineRunStatus.SUCCESS,
            )
        )
        storage.add_run(
            TestRunStorage.build_run(
                run_id=two, pipeline_name='some_pipeline', status=PipelineRunStatus.SUCCESS,
            ),
        )

        assert len(storage.get_runs()) == 2

        some_runs = storage.get_runs(PipelineRunsFilter(run_ids=[one, two]))
        count = storage.get_runs_count(PipelineRunsFilter(run_ids=[one, two]))
        assert len(some_runs) == 2
        assert count == 2

    def test_fetch_run_group(self, storage):
        assert storage
        root_run = TestRunStorage.build_run(run_id=make_new_run_id(), pipeline_name='foo_pipeline')
        runs = [root_run]

        # Create 3 children and 3 descendants of the rightmost child:
        #    root
        #   /  |  \
        # [0] [1] [2]
        #          |
        #         [a]
        #          |
        #         [b]
        #          |
        #         [c]

        for _ in range(3):
            runs.append(
                TestRunStorage.build_run(
                    run_id=make_new_run_id(),
                    pipeline_name='foo_pipeline',
                    root_run_id=root_run.run_id,
                    parent_run_id=root_run.run_id,
                    tags={PARENT_RUN_ID_TAG: root_run.run_id, ROOT_RUN_ID_TAG: root_run.run_id},
                )
            )
        for _ in range(3):
            # get root run id from the previous run if exists, otherwise use previous run's id
            root_run_id = runs[-1].root_run_id if runs[-1].root_run_id else runs[-1].run_id
            parent_run_id = runs[-1].run_id
            runs.append(
                TestRunStorage.build_run(
                    run_id=make_new_run_id(),
                    pipeline_name='foo_pipeline',
                    root_run_id=root_run_id,
                    parent_run_id=parent_run_id,
                    tags={PARENT_RUN_ID_TAG: parent_run_id, ROOT_RUN_ID_TAG: root_run_id},
                )
            )
        for run in runs:
            storage.add_run(run)

        run_group_one = storage.get_run_group(root_run.run_id)

        assert len(run_group_one[1]) == 7

        run_group_two = storage.get_run_group(runs[-1].run_id)

        assert len(run_group_two[1]) == 7

        assert run_group_one[0] == run_group_two[0]
        assert run_group_one[1] == run_group_two[1]

    def test_fetch_run_group_not_found(self, storage):
        assert storage
        run = TestRunStorage.build_run(run_id=make_new_run_id(), pipeline_name='foo_pipeline')
        storage.add_run(run)

        run_group_result = storage.get_run_group(make_new_run_id())
        assert run_group_result is None

    def test_fetch_run_groups(self, storage):
        assert storage
        root_runs = [
            TestRunStorage.build_run(run_id=make_new_run_id(), pipeline_name='foo_pipeline')
            for i in range(3)
        ]
        runs = [run for run in root_runs]
        for _ in range(5):
            for root_run in root_runs:
                runs.append(
                    TestRunStorage.build_run(
                        run_id=make_new_run_id(),
                        pipeline_name='foo_pipeline',
                        tags={PARENT_RUN_ID_TAG: root_run.run_id, ROOT_RUN_ID_TAG: root_run.run_id},
                    )
                )
        for run in runs:
            storage.add_run(run)

        run_groups = storage.get_run_groups(limit=5)

        assert len(run_groups) == 3

        expected_group_lens = {
            root_runs[i].run_id: expected_len for i, expected_len in enumerate([2, 3, 3])
        }

        for root_run_id in run_groups:
            assert len(run_groups[root_run_id]['runs']) == expected_group_lens[root_run_id]
            assert run_groups[root_run_id]['count'] == 6

    def test_fetch_run_groups_filter(self, storage):
        assert storage

        root_runs = [
            TestRunStorage.build_run(run_id=make_new_run_id(), pipeline_name='foo_pipeline')
            for i in range(3)
        ]

        runs = [run for run in root_runs]
        for root_run in root_runs:
            failed_run_id = make_new_run_id()
            runs.append(
                TestRunStorage.build_run(
                    run_id=failed_run_id,
                    pipeline_name='foo_pipeline',
                    tags={PARENT_RUN_ID_TAG: root_run.run_id, ROOT_RUN_ID_TAG: root_run.run_id},
                    status=PipelineRunStatus.FAILURE,
                )
            )
            for _ in range(3):
                runs.append(
                    TestRunStorage.build_run(
                        run_id=make_new_run_id(),
                        pipeline_name='foo_pipeline',
                        tags={PARENT_RUN_ID_TAG: failed_run_id, ROOT_RUN_ID_TAG: root_run.run_id},
                    )
                )

        for run in runs:
            storage.add_run(run)

        run_groups = storage.get_run_groups(
            limit=5, filters=PipelineRunsFilter(status=PipelineRunStatus.FAILURE)
        )

        assert len(run_groups) == 3

        for root_run_id in run_groups:
            assert len(run_groups[root_run_id]['runs']) == 2
            assert run_groups[root_run_id]['count'] == 5

    def test_fetch_run_groups_ordering(self, storage):
        assert storage

        first_root_run = TestRunStorage.build_run(
            run_id=make_new_run_id(), pipeline_name='foo_pipeline'
        )

        storage.add_run(first_root_run)

        second_root_run = TestRunStorage.build_run(
            run_id=make_new_run_id(), pipeline_name='foo_pipeline'
        )

        storage.add_run(second_root_run)

        second_root_run_child = TestRunStorage.build_run(
            run_id=make_new_run_id(),
            pipeline_name='foo_pipeline',
            tags={
                PARENT_RUN_ID_TAG: second_root_run.run_id,
                ROOT_RUN_ID_TAG: second_root_run.run_id,
            },
        )

        storage.add_run(second_root_run_child)

        first_root_run_child = TestRunStorage.build_run(
            run_id=make_new_run_id(),
            pipeline_name='foo_pipeline',
            tags={
                PARENT_RUN_ID_TAG: first_root_run.run_id,
                ROOT_RUN_ID_TAG: first_root_run.run_id,
            },
        )

        storage.add_run(first_root_run_child)

        run_groups = storage.get_run_groups(limit=1)

        assert first_root_run.run_id in run_groups
        assert second_root_run.run_id not in run_groups



from collections import defaultdict

from dagster import (
    DependencyDefinition,
    InputDefinition,
    Int,
    PipelineDefinition,
    SolidInvocation,
    execute_pipeline,
    lambda_solid,
    solid,
)


def test_aliased_solids():
    @lambda_solid()
    def first():
        return ['first']

    @lambda_solid(input_defs=[InputDefinition(name="prev")])
    def not_first(prev):
        return prev + ['not_first']

    pipeline = PipelineDefinition(
        solid_defs=[first, not_first],
        dependencies={
            'not_first': {'prev': DependencyDefinition('first')},
            SolidInvocation('not_first', alias='second'): {
                'prev': DependencyDefinition('not_first')
            },
            SolidInvocation('not_first', alias='third'): {'prev': DependencyDefinition('second')},
        },
    )

    result = execute_pipeline(pipeline)
    assert result.success
    solid_result = result.result_for_solid('third')
    assert solid_result.output_value() == ['first', 'not_first', 'not_first', 'not_first']


def test_only_aliased_solids():
    @lambda_solid()
    def first():
        return ['first']

    @lambda_solid(input_defs=[InputDefinition(name="prev")])
    def not_first(prev):
        return prev + ['not_first']

    pipeline = PipelineDefinition(
        solid_defs=[first, not_first],
        dependencies={
            SolidInvocation('first', alias='the_root'): {},
            SolidInvocation('not_first', alias='the_consequence'): {
                'prev': DependencyDefinition('the_root')
            },
        },
    )

    result = execute_pipeline(pipeline)
    assert result.success
    solid_result = result.result_for_solid('the_consequence')
    assert solid_result.output_value() == ['first', 'not_first']


def test_aliased_configs():
    @solid(input_defs=[], config_schema=Int)
    def load_constant(context):
        return context.solid_config

    pipeline = PipelineDefinition(
        solid_defs=[load_constant],
        dependencies={
            SolidInvocation(load_constant.name, 'load_a'): {},
            SolidInvocation(load_constant.name, 'load_b'): {},
        },
    )

    result = execute_pipeline(
        pipeline, {'solids': {'load_a': {'config': 2}, 'load_b': {'config': 3}}}
    )

    assert result.success
    assert result.result_for_solid('load_a').output_value() == 2
    assert result.result_for_solid('load_b').output_value() == 3


def test_aliased_solids_context():
    record = defaultdict(set)

    @solid
    def log_things(context):
        solid_value = context.solid.name
        solid_def_value = context.solid_def.name
        record[solid_def_value].add(solid_value)

    pipeline = PipelineDefinition(
        solid_defs=[log_things],
        dependencies={
            SolidInvocation('log_things', 'log_a'): {},
            SolidInvocation('log_things', 'log_b'): {},
        },
    )

    result = execute_pipeline(pipeline)
    assert result.success

    assert dict(record) == {'log_things': set(['log_a', 'log_b'])}
